'''
    Automates VCF file filering using VCFtools. Filters may be used independently or
    combined to create complex filtering operations.        

    Please note, `VCFtools <https://vcftools.github.io/index.html>`_ should be cited
    alonside the PPP when using this function. Please see the log file for the relevant 
    citation.

    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file to be filtered.
    **--model-file** *<model_filename>*
    	Argument used to define the model file. Please note that this argument cannot be 
    	used with the individual-based filters.
    **--model** *<model_str>*
    	Argument used to define the model (i.e. the individual(s) to include). Please note
    	that this argument cannot be used with the individual-based filters.

    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-format** *<vcf, vcf.gz, bcf, removed_sites, kept_sites, removed_bed, kept_bed>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; BCF (bcf); removed or kept variants
        in sites format (removed_sites, kept_sites); or removed or kept variants in bed
        format (removed_bed, kept_bed).
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    #############################
    Filter Command-line Arguments
    #############################
    The filtering arguments below are roughly seperated into catagoires. Please not that
    mulitple filters are seperated into two opposing function types **include** and 
    **exclude**.

    **************************
    Individual-Based Arguments
    **************************
    Please note that all individual-based arguments are not compatible with either the 
    **--model** or **--model-file** command-line arguments.

    **--filter-include-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to include. This argument may be used 
        multiple times if desired.
    **--filter-exclude-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to exclude. This argument may be used 
        multiple times if desired.
    **--filter-include-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to include.
    **--filter-exclude-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to exclude.

    *******************************
    Allele/Genotype-Based Arguments
    *******************************
    **--filter-min-alleles** *min_int*
    	Argument used to include positions with a number of allele >= to the 
    	given number.
    **--filter-max-alleles** *max_int*
        Argument used to include positions with a number of allele <= to the 
        given number.
    **--filter-maf-min** *maf_proportion*
    	Argument used to include sites with equal or greater MAF values.
    **--filter-maf-max** *maf_proportion*
    	Argument used to include sites with equal or lesser MAF values.
    **--filter-mac-min** *mac_int*
    	Argument used to include sites with equal or greater MAC values.
    **--filter-mac-max** *mac_int*
    	Argument used to include sites with equal or lesser MAC values.
    	**--filter-include-indels**
    	Argument used to include positions if they contain an insertion or a deletion. 
    **--filter-exclude-indels**
    	Argument used to exclude positions if they contain an insertion or a deletion. 
    **--filter-include-snp** *<rs#>* *<rs#1, rs#2, etc.>*
    	Argument used to include SNP(s) with the matching ID. This argument may be used 
        multiple times if desired.
    **--filter-include-snp-file** *<snp_filename>*
    	Argument used to define a file of SNP IDs to include.
    **--filter-exclude-snp-file** *<snp_filename>*
    	Argument used to define a file of SNP IDs to exclude.
    **--filter-max-missing** *proportion_float*
    	Argument used to filter positions by their proportion of missing data, a value of 
    	0.0 will ignore missing data whereas a value of 1.0 will allow no missing data.

    ************************
    Position-Based Arguments
    ************************
    **--filter-include-chr** *<chr>* *<chr1, chr2, etc.>*
        Argument used to define the chromosome(s) to include. This argument may be used 
        multiple times if desired.
    **--filter-exclude-chr** *<chr>* *<chr1, chr2, etc.>*
        Argument used to define the chromosome(s) to exclude. This argument may be used 
        multiple times if desired.
    **--filter-from-bp** *<bp_int>*
        Argument used to define the lower bound of positions to include. May only be used 
        with a single chromosome.
    **--filter-to-bp** *<bp_int>*
        Argument used to define the upper bound of positions to include. May only be used 
        with a single chromosome.
    **--filter-distance** *<bp_int>*
    	Argument used to define the distance that no two sites may be within
    **--filter-include-positions** *<position_filename>*
        Argument used to define a file of positions to include within a tsv file 
        (chromosome and position).
    **--filter-exclude-positions** *<position_filename>*
        Argument used to define a file of positions to exclude within a tsv file 
        (chromosome and position).
    **--filter-include-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to include.
    **--filter-exclude-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to exclude.

    ********************
    Flag-Based Arguments
    ********************
    **--filter-include-passed**
        Argument used to include positions with the 'PASS' filter flag.
    **--filter-include-filtered** *<filter_flag>*
        Argument used to include positions with the specified filter flag.
    **--filter-exclude-filtered** *<filter_flag>*
        Argument used to exclude positions with the specified filter flag.
    **--filter-include-info** *<info_flag>*
        Argument used to include positions with the specified info flag.
    **--filter-exclude-info** *<info_flag>*
        Argument used to exclude positions with the specified info flag.

    ##########################
    Example Command-line Usage
    ##########################
    Command-line to only include biallelic sites:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.bcf --filter-min-alleles 2 --filter-max-alleles 2

    Command-line to only include chromosome chr1:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.vcf.gz --filter-include-chr chr1

    Command-line to remove indels and ouput a BCF file:

    .. code-block:: bash
        
        python vcf_filter.py --vcf input.vcf --filter-exclude-indels --out-format bcf
'''

import os
import sys
import subprocess
import argparse
import logging

# Import basic vcftools functions
from pgpipe.vcftools import *

# Model file related functions
from pgpipe.model import read_model_file

# Import logging functions
from pgpipe.logging_module import initLogger, logArgs

def vcf_filter_parser(passed_arguments):
    '''
    VCF Filter Argument Parser

    Assign the parameters for the VCF Filter using argparse.

    Parameters
    ----------
	passed_arguments : list, optional
		Parameters passed by another function. sys.argv is used if
		not given. 

	Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]
                
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())

    # Model file arguments.
    vcf_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    vcf_parser.add_argument('--model', help = 'Defines the model and the individual(s) to include', type = str)

    # Other file arguments. Expand as needed
    vcf_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix', type = str)
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', type = str,  default = 'out')
    #vcf_parser.add_argument('--log', help = "Specifies if the vcftools log should be saved", action = 'store_false')

    out_format_list = ['vcf', 'vcf.gz', 'bcf', 'removed_sites', 'kept_sites', 'removed_bed', 'kept_bed']
    out_format_default = 'vcf.gz'

    vcf_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)

    # General arguments.
    vcf_parser.add_argument('--overwrite', help = "Overwrite previous output", action = 'store_true')

    # Galaxy Option to pipe log to stdout
    vcf_parser.add_argument('--log-stdout', help = argparse.SUPPRESS, action = 'store_true')

    ### Filters

    # Non-model file arguments
    vcf_parser.add_argument('--filter-include-indv', help = 'Defines the individual(s) to include. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-indv', help = 'Defines the individual(s) to exclude. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-include-indv-file', help = 'Defines a file of individuals to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-indv-file', help = 'Defines a file of individuals to exclude', action = parser_confirm_file())

    # Allele count filters
    vcf_parser.add_argument('--filter-min-alleles', help = 'Include positions with a number of allele >= to the given number', type = int)
    vcf_parser.add_argument('--filter-max-alleles', help = 'Include positions with a number of allele <= to the given number', type = int)

    # Missing data filter
    vcf_parser.add_argument('--filter-max-missing', help = 'Filter positions by their proportion of missing data (0.0: include all, 1.0: no missing data)', type = float)

    # Indel filters
    indel_filters = vcf_parser.add_mutually_exclusive_group()
    indel_filters.add_argument('--filter-include-indels', help = 'Include positions if they contain an insertion or a deletion', action = 'store_true')
    indel_filters.add_argument('--filter-exclude-indels', help = 'Exclude positions if they contain an insertion or a deletion', action = 'store_true')

    # Chromosome filters
    vcf_parser.add_argument('--filter-include-chr', help = 'The chromosome(s) to include. This argument may be used multiple times', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-chr', help = 'The chromosome(s) to exclude. This argument may be used multiple times', nargs = '+', type = str, action = parser_add_to_list())

    # Basic position filters
    vcf_parser.add_argument('--filter-from-bp', help = 'Defines the lower bound of positions to include. May only be used with a single chromosome', type = int)
    vcf_parser.add_argument('--filter-to-bp', help = 'Defines the upper bound of positions to include. May only be used with a single chromosome', type = int)

    # Position-based position filters
    vcf_parser.add_argument('--filter-include-positions', help = 'Defines a file of positions to include within a tsv file (chromosome and position)', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-positions', help = 'Defines a file of positions to exclude within a tsv file (chromosome and position)', action = parser_confirm_file())

    # BED-based position filters
    vcf_parser.add_argument('--filter-include-bed', help = 'Defines a BED file of positions to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-bed', help = 'Defines a BED file of positions to exclude', action = parser_confirm_file())

    # Filter-flag filters
    vcf_parser.add_argument('--filter-include-passed', help = "Include positions with the 'PASS' filter flag", action = 'store_true')
    vcf_parser.add_argument('--filter-include-flag', help = 'Include positions with the specified filter flag', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-flag', help = 'Exclude positions with the specified filter flag', nargs = '+', type = str, action = parser_add_to_list())

    # Info-flag filters
    vcf_parser.add_argument('--filter-include-info', help = 'Include positions with the specified info flag', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-info', help = 'Exclude positions with the specified info flag', nargs = '+', type = str, action = parser_add_to_list())

    # SNP filters
    vcf_parser.add_argument('--filter-include-snp', help = 'Include SNP(s) with the matching ID. This argument may be used multiple times.', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-include-snp-file', help = 'Defines a file of SNP IDs to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-snp-file', help = 'Defines a file of SNP IDs to exclude', action = parser_confirm_file())

    # MAF Filter
    vcf_parser.add_argument('--filter-maf-min', help = 'Include sites with equal or greater MAF values', type = float)
    vcf_parser.add_argument('--filter-maf-max', help = 'Include sites with equal or lesser MAF values', type = float)

    # MAC Filter
    vcf_parser.add_argument('--filter-mac-min', help = 'Include sites with equal or greater MAC values', type = int)
    vcf_parser.add_argument('--filter-mac-max', help = 'Include sites with equal or lesser MAC values', type = int)

    # Additional Filters
    vcf_parser.add_argument('--filter-distance', help = 'Distance that no two sites may be within', type = int)

    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()

def run (passed_arguments = []):
    '''
    Filter function for the PPP

    This function uses the argparse-based function :py:func:`vcf_filter_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their VCFtools equivalent. Once all the
    VCFtools-based parameters are assigned, VCFtools is called.

    Parameters
    ----------
    --vcf : str
        Input VCF filename
    --out : str
        Complete output filename, overrides --out-prefix
    --out-prefix : str
        Output filename prefix
    --out-format : str
        Output format
	--model-file : str
		Model filename
	--model : str
		Model to use
    --filter-include-indv : list or str
        Individual(s) to include. May be used multiple times. Not usable w/ --model
    --filter-exclude-indv : list or str
        Individual(s) to exclude. May be used multiple times. Not usable w/ --model
    --filter-include-indv-file : str
        File of individuals to include. Not usable w/ --model
    --filter-exclude-indv-file : str
        File of individuals to exclude.  Not usable w/ --model
    --filter-min-alleles : int
        Include positions with a number of allele >= to the given number
    --filter-max-alleles : int
        Include positions with a number of allele <= to the given number
    --filter-maf-min
		Include sites with equal or greater MAF values
    --filter-maf-max
    	Include sites with equal or lesser MAF values
    --filter-mac-min
    	Include sites with equal or greater MAC values
    --filter-mac-max
    	Include sites with equal or lesser MAC values
   	--filter-include-indels
    	Include positions if they contain an insertion or a deletion
    --filter-exclude-indels
    	Exclude positions if they contain an insertion or a deletion
 	--filter-include-snp
 		Include SNP(s) with the matching ID. May be used multiple times
    --filter-include-snp-file
    	File of SNP IDs to include
    --filter-exclude-snp-file
    	File of SNP IDs to exclude
    --filter-max-missing
		Filter positions by their proportion of missing data. (0.0: include all, 1.0: no missing data)
    --filter-include-chr
        Chromosome(s) to include. May be used multiple times.
    --filter-exclude-chr
        Chromosome(s) to exclude. May be used multiple times.
    --filter-from-bp : int
        Lower bound of positions to include. Only usable with a single chromosome
    --filter-to-bp : int
        Upper bound of positions to include. Only usable with a single chromosome
    --filter-include-positions : str
        File of positions to include (tsv: chromosome and position)
    --filter-exclude-positions : str
        File of positions to exclude (tsv: chromosome and position)
    --filter-include-bed : str
        BED file of positions to include 
    --filter-exclude-bed : str
        BED file of positions to exclude
    --filter-include-passed : bool
        Include positions with the 'PASS' filter flag
    --filter-include-filtered : list or str
        Include positions with the specified filter flag
    --filter-exclude-filtered : list or str
        Exclude positions with the specified filter flag
    --filter-include-info : list or str
        Include positions with the specified info flag
    --filter-exclude-info : list or str
        Exclude positions with the specified info flag

    Raises
    ------
    IOError
        Output file already exists and --overwrite is not specified
    Exception
        Incompatible arguments

    '''

    # Grab VCF arguments from command line
    vcf_args = vcf_filter_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_filter')

    # Argument container for vcftools
    vcftools_call_args = []

    # Check if the user has specified a model file
    if vcf_args.model_file and vcf_args.model:
        # Read in the models
        models_in_file = read_model_file(vcf_args.model_file)

        # Check that the selected model was not found in the file
        if vcf_args.model not in models_in_file:
            raise IOError('Selected model "%s" not found in: %s' % (vcf_args.model, vcf_args.model_file))

        # Check that individual-based filters are not also being used
        if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
            if vcf_args.filter_include_indv:
                raise Exception('--model and --filter-include-indv arguments are incompatible')
            if vcf_args.filter_exclude_indv:
                raise Exception('--model and --filter-exclude-indv arguments are incompatible')

        # Check that individuals-based filters are not also being used
        if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
            if vcf_args.filter_include_indv_file:
                raise Exception('--model and --filter_include_indv-file arguments are incompatible')
            if vcf_args.filter_exclude_indv_file:
                raise Exception('--model and --filter-exclude-indv-file arguments are incompatible')

        # Select model, might change this in future versions
        selected_model = models_in_file[vcf_args.model]

        # Create individuals file
        selected_model.create_ind_file(overwrite = vcf_args.overwrite)

        # Assign the individuals file to vcftools
        vcftools_call_args.extend(['--keep', selected_model.ind_file])

    # Holds the filename suffix vcftools assigns to the filtered output
    vcftools_out_suffix = None

    # Used to assign the output format to the vcftools call, and assign vcftools output suffix
    if vcf_args.out_format == 'removed_sites':
        vcftools_call_args.append('--removed-sites')
        vcftools_out_suffix = '.removed.sites'
    elif vcf_args.out_format == 'kept_sites':
        vcftools_call_args.append('--kept-sites')
        vcftools_out_suffix = '.kept.sites'
    elif vcf_args.out_format == 'removed_bed':
        vcftools_call_args.append('--removed-sites')
        vcftools_out_suffix = '.removed.bed'
    elif vcf_args.out_format == 'kept_bed':
        vcftools_call_args.append('--kept-sites')
        vcftools_out_suffix = '.kept.bed'
    elif vcf_args.out_format == 'bcf':
        vcftools_call_args.append('--recode')
        vcftools_out_suffix = '.recode.bcf'
    elif vcf_args.out_format == 'vcf':
        vcftools_call_args.append('--recode')
        vcftools_out_suffix = '.recode.vcf'
    elif vcf_args.out_format == 'vcf.gz':
        vcftools_call_args.append('--recode')
        vcftools_out_suffix = '.recode.vcf.gz'

    # Assign expected vcftools output filename
    vcftools_output_filename = vcf_args.out_prefix + vcftools_out_suffix

    # Check if the user has specified a output filename
    if vcf_args.out:

        # Rename the vcftools output filename, using the output filename
        vcftools_output_filename = vcf_args.out

    # Check if previous output should be overwritten
    if not vcf_args.overwrite:

        # Confirm the vcftools output and log file do not exist
        check_for_vcftools_output(vcftools_output_filename)

    # Individuals-based filters
    if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
        # Used to include a file of individuals to keep
        if vcf_args.filter_include_indv_file:
            vcftools_call_args.extend(['--keep', vcf_args.filter_include_indv_file])

        # Used to include a file of individuals to remove
        if vcf_args.filter_exclude_indv_file:
            vcftools_call_args.extend(['--remove', vcf_args.filter_exclude_indv_file])

    # Individual-based filters
    if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
        if vcf_args.filter_include_indv:
            for indv_to_include in vcf_args.filter_include_indv:
                vcftools_call_args.extend(['--indv', indv_to_include])
        if vcf_args.filter_exclude_indv:
            for indv_to_exclude in vcf_args.filter_exclude_indv:
                vcftools_call_args.extend(['--remove-indv', indv_to_exclude])

    # Indel-based filters
    if vcf_args.filter_include_indels or vcf_args.filter_exclude_indels:
        if vcf_args.filter_include_indels:
            vcftools_call_args.append('--keep-only-indels')
        if vcf_args.filter_exclude_indels:
            vcftools_call_args.append('--remove-indels')

    # Chromosome-based filters
    if vcf_args.filter_include_chr or vcf_args.filter_exclude_chr:
        if vcf_args.filter_include_chr:
            for chr_to_include in vcf_args.filter_include_chr:
                vcftools_call_args.extend(['--chr', chr_to_include])
        if vcf_args.filter_exclude_chr:
            for chr_to_exclude in vcf_args.filter_exclude_chr:
                vcftools_call_args.extend(['--not-chr', chr_to_exclude])

    # Site (i.e. basepair) filters
    if vcf_args.filter_from_bp or vcf_args.filter_to_bp:
        if vcf_args.filter_include_chr:
            vcftools_call_args.extend(['--from-bp', vcf_args.filter_from_bp])
        if vcf_args.filter_exclude_chr:
            vcftools_call_args.extend(['--to-bp', vcf_args.filter_to_bp])

    # Position (vcftools output file) filters
    if vcf_args.filter_include_positions or vcf_args.filter_exclude_positions:
        if vcf_args.filter_include_positions:
            vcftools_call_args.extend(['--positions', vcf_args.filter_include_positions])
        if vcf_args.filter_exclude_positions:
            vcftools_call_args.extend(['--exclude-positions', vcf_args.filter_exclude_positions])

    # Position (BED format file) filters
    if vcf_args.filter_include_bed or vcf_args.filter_exclude_bed:
        if vcf_args.filter_include_bed:
            vcftools_call_args.extend(['--bed', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            vcftools_call_args.extend(['--exclude-bed', vcf_args.filter_exclude_bed])

    # Flag-based Filters
    if vcf_args.filter_include_passed or vcf_args.filter_include_flag or vcf_args.filter_exclude_flag:
        if vcf_args.filter_include_passed:
            vcftools_call_args.append('--remove-filtered-all')
        if vcf_args.filter_include_flag:
            for filtered_to_include in vcf_args.filter_include_flag:
                vcftools_call_args.extend(['--keep-filtered', filtered_to_include])
        if vcf_args.filter_exclude_flag:
            for filtered_to_exclude in vcf_args.filter_exclude_flag:
                vcftools_call_args.extend(['--remove-filtered', filtered_to_exclude])

    # Infor-based filters
    if vcf_args.filter_include_info or vcf_args.filter_exclude_info:
        if vcf_args.filter_include_info:
            for info_to_include in vcf_args.filter_include_info:
                vcftools_call_args.extend(['--keep-INFO', info_to_include])
        if vcf_args.filter_exclude_info:
            for info_to_exclude in vcf_args.filter_exclude_info:
                vcftools_call_args.extend(['--remove-INFO', info_to_exclude])

    # Allele-based filters
    if vcf_args.filter_min_alleles or vcf_args.filter_max_alleles:
        if vcf_args.filter_min_alleles:
            vcftools_call_args.extend(['--min-alleles', vcf_args.filter_min_alleles])
        if vcf_args.filter_max_alleles:
            vcftools_call_args.extend(['--max-alleles', vcf_args.filter_max_alleles])

    # Missing data filters
    if vcf_args.filter_max_missing:
        vcftools_call_args.extend(['--max-missing', vcf_args.filter_max_missing])

    # SNP-based filters
    if vcf_args.filter_include_snp or vcf_args.filter_include_snp_file or vcf_args.filter_exclude_snp_file:
        if vcf_args.filter_include_snp:
            for snp_to_include in vcf_args.filter_include_snp:
                vcftools_call_args.extend(['--snp', snp_to_include])
        if vcf_args.filter_include_snp_file:
            vcftools_call_args.extend(['--snps', vcf_args.filter_include_snp_file])
        if vcf_args.filter_exclude_snp_file:
            vcftools_call_args.extend(['--exclude', vcf_args.filter_exclude_snp_file])

    # MAF-based filters
    if vcf_args.filter_maf_min or vcf_args.filter_maf_max:
        if vcf_args.filter_maf_min:
            vcftools_call_args.extend(['--maf', vcf_args.filter_maf_min])
        if vcf_args.filter_maf_max:
            vcftools_call_args.extend(['--max-maf', vcf_args.filter_maf_max])

    # MAC-based filters
    if vcf_args.filter_mac_min or vcf_args.filter_mac_max:
        if vcf_args.filter_mac_min:
            vcftools_call_args.extend(['--mac', vcf_args.filter_mac_min])
        if vcf_args.filter_mac_max:
            vcftools_call_args.extend(['--max-mac', vcf_args.filter_mac_max])

    # Distance (between sites) filters
    if vcf_args.filter_distance:
        vcftools_call_args.extend(['--thin', vcf_args.filter_distance])

    logging.info('vcftools parameters assigned')

    # Assigns the file argument for vcftools
    vcfname_arg = assign_vcftools_input_arg(vcf_args.vcf)
    logging.info('Input file assigned')

    # Check if the output format is vcf
    if vcf_args.out_format == 'vcf':

        # Call vcftools with the specifed arguments
        vcftools_err = call_vcftools(vcfname_arg + vcftools_call_args, output_format = vcf_args.out_format, output_filename = vcftools_output_filename)

    else:

        # Call vcftools with the specifed arguments
        vcftools_err = call_vcftools(vcfname_arg + vcftools_call_args, output_format = vcf_args.out_format, output_filename = vcftools_output_filename)

    # Check if the log should be piped to the stdout
    if vcf_args.log_stdout:

        # Write the log to stdout
        sys.stdout.write(vcftools_err)

    # Check if log should be saved as a file
    else:

        # Create the log file
        produce_vcftools_log(vcftools_err, vcftools_output_filename)

    # Delete any files that were created for vcftools
    if vcf_args.model_file:
        selected_model.delete_ind_file()

if __name__ == "__main__":
    initLogger()
    run()



#!/usr/bin/env python
'''
    Wrapper for IMa3 executable. This wrapper will handle locating the proper
    version of IMa3 for the desired run, as well as providing multi-threading
    support if the system has the mpirun utility and the IMa executable
    was compiled with it. Most options for IMa3 are described in the IMa3 
    manual, and it is HIGHLY recommended that users overview this document
    before calling IMa3. Note that wrapper functions (--threads,--ima-path)
    require two dashes, while IMa arguments use only one.

    ###############
    Input Arguments
    ###############
    **-i** *<input_file>*
        Name of IMa3 input file generated by PPP
    **-o** *<output_file>*
        Name of IMa3 output file. Additional files will use this as prefix.

    ###################
    Parameter Arguments
    ###################
    **-q** *<max_pop_size>*
        Sets maximum population size parameter for all populations.
    **-m** *<migration_rate>*
        Sets migration rate prior.
    **-t** *<max_split_time>*
        Sets maximum splitting time parameter.

    #################
    Wrapper Arguments
    #################
    **--threads** *<thread_count>*
        Set number of threads to use. This will check that the proper version
        of IMa3 has been compiled and the system has mpirun installed.
    **--ima-path** *<path_to_ima>*
        Path to IMa executable to use if not on system path. This should include
        the name of the executable, not just the path to it.
'''

import os
import sys
import subprocess
import argparse
import logging
import itertools
from distutils.spawn import find_executable


#from logging_module import initLogger

def imaParser():
    '''Ima2p Argument Parser - Assigns arguments from command line'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError # File not found
                setattr(args, self.dest, value)
        return customAction

    parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    parser.add_argument("-i", metavar = 'Ima2p_Input', help = "Specifies the name of input data file", type = str)
    parser.add_argument("-o", help = "Specifies the name for program output files", type = str)
    parser.add_argument("-b", help = "Specifies the duration of burn : number of burnin steps")
    parser.add_argument("-c", help = "Specifies the calculation options", type = int)
    parser.add_argument("-d", help = "Specifies the number of steps between tree saving", type = int)
    parser.add_argument("-f", help = "Specifies the name of file with saved Markov chain state generated in previous run -use with -r3", type = str)
    parser.add_argument("-g", help = "Specifies the name of file with parameter priors  (requires -c3)", type = str)
    parser.add_argument("-hf", help = "Specifies the heating model : geometric", type = str) 
    parser.add_argument("-hn", help = "Specifies the number of chains", type = int)
    parser.add_argument("-hk", help = "Specifies the number of chain swap attempts per step", type = int)
    parser.add_argument("-ha", help = "Specifies the first heating parameter, effect depends on heating model", type = float)
    parser.add_argument("-hb", help = "Specifies the second heating parameter, effect depends on heating model", type = float)
    parser.add_argument("-j", help = "Specifies the model options", type = int)
    parser.add_argument("-l", help = "Specifies the run duration", type = int)
    parser.add_argument("-m", help = "Specifies the migration prior value", type = float)
    parser.add_argument("-p", help = "Specifies the output options", type = int)
    parser.add_argument("-q", help = "Specifies the maximum for population size parameters", type = int)
    parser.add_argument("-r", help = "Specifies the run options", type = int)
    parser.add_argument("-s", help = "Specifies the random number seed", type = int)
    parser.add_argument("-t", help = "Specifies the maximum time of population splitting", type = int)
    parser.add_argument("-u", help = "Specifies the Generation time in years -for use with -p3", type = int)
    parser.add_argument("-v", help = "Specifies the base name", type = str)
    parser.add_argument("-w", help = "Specifies the name of file with nested models to be tested (LOAD-GENEALOGY mode only), invokes -c2)", type = str)
    parser.add_argument("-y", help = "Mutation rate scalar for relevant loci - for use with -p3")
    parser.add_argument("-z", help = "Specifies the number of stepsbetween screen output", type = int)
    return parser

def wrapperParser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--threads",type=int,help="Use multi-threaded mode with this thread count")
    parser.add_argument("--ima-path",type=str,help="Path to IMa3 exe if not on system path")
    return parser

def bothParser(p1,p2):
    parser = argparse.ArgumentParser(parents=[p1,p2])
    return parser


def getIMaMT():
    for exe in ['IMa3','IMa3_xml']:
        if find_executable(exe) is not None:
            return exe
    return None

def getIMaST():
    for exe in ['IMa3_singlecpu','IMa3_xml_singlecpu']:
        if find_executable(exe) is not None:
            return exe
    return None

def getIMaExe(args):
    if args.ima_path is not None:
        if not os.path.isfile(args.ima_path):
            raise Exception("IMa exe at %s does not exist"%args.ima_path)
        return args.ima_path
    imamt = getIMaMT()
    if args.threads is not None:
        return imamt
    if imamt is not None:
        return imamt
    return getIMaST()
    


def call_ima3(args,ima_arglist):
    run_args = []
    #if find_executable('mpirun') is not None:
    if args.threads is not None:
        if find_executable('mpirun') is None:
            raise Exception('mpirun required on path for multi-threading')
        run_args.extend(['mpirun','-np',str(args.threads)])
    ima_exe = getIMaExe(args)
    if ima_exe is None:
        raise Exception("No IMa executable found on system path")
    run_args.append(ima_exe)
    run_args.extend(map(str,ima_arglist))
    ima_call = subprocess.Popen(run_args,stdout=subprocess.PIPE,stderr=subprocess.PIPE)
    ima_stdout, ima_stderr = ima_call.communicate()
    if sys.version_info[0] == 3:
        ima_stdout = ima_stdout.decode()
        ima_stderr = ima_stderr.decode()
    if ima_call.returncode != 0:
        raise Exception("Error with IMa call:%d\n%s"%(ima_call.returncode,ima_stderr))
    #sys.stdout.write(ima_stdout+'\n')
    sys.stdout.write(ima_stderr+'\n')
    


def logArgs(args, pipeline_function):
    '''Logs arguments from argparse system. May be replaced with a general function in logging_module'''
    logging.info('Arguments for %s:' % pipeline_function)
    for k in vars(args):
        logging.info('Arguments %s: %s' % (k, vars(args)[k]))

def run (sysargv):
    
    ima_parser = imaParser()
    misc_parser = wrapperParser()
    main_parser = bothParser(ima_parser,misc_parser)
    if len(sysargv) == 0:
        main_parser.print_help()
        exit()
    misc_args, ima_arglist = misc_parser.parse_known_args(sysargv)

    #Check that arguments don't cause error, since list is passed to 
    #popen the returned namespace isn't used
    ima_parser.parse_known_args(ima_arglist)
    call_ima3(misc_args,ima_arglist)
    logging.info("Run successful")


if __name__ == "__main__":
    #initLogger()
    run(sys.argv[1:])





import sys
from pgpipe.vcf_reader_func import VcfReader, checkPopWithoutMissing, readSinglePopModel
import pysam



vcfname = str(sys.argv[1])
popmodel = readSinglePopModel(sys.argv[2])
vcfr = VcfReader(vcfname,popmodel = popmodel)
rec_list = vcfr.getRecordList()

print (checkPopWithoutMissing(rec_list,popmodel,vcfr.popkeys))




#!/usr/bin/env python
'''
    Automates the calculation of multiple LD statistics using Plink.

    ############################
    Input Command-line Arguments
    ############################
    **--ped-prefix** *<input_prefix>*
        Argument used to define the filename prefix shared by the ped file (.ped) and
        the map file (.map). Should not be used alongside the specific file arguments 
        (e.g. --ped).
    **--ped** *<ped_filename>*
        Argument used to define the filename of the plink ped file (.ped). Must be 
        called alongside --map. Cannot be called alongside --ped-prefix.
    **--map** *<map_filename>*
        Argument used to define the filename of the plink map file (.map). Must be 
        called alongside --ped. Cannot be called alongside --ped-prefix.
    **--binary-ped-prefix** *<input_prefix>*
        Argument used to define the filename prefix shared by the binary ped file
        (.bed), the fam file (.fam), and the bim file (.bim). Should not be used 
        alongside the specific file arguments (e.g. --binary-ped).
    **--binary-ped** *<binary_ped_filename>*
        Argument used to define the filename of the plink binary ped file (.bed). 
        Must be called alongside --fam and --bim. Cannot be called alongside 
        --binary-ped-prefix.
    **--fam** *<fam_filename>*
        Argument used to define the filename of the plink fam file (.fam). Must be 
        called alongside --binary-ped and --bim. Cannot be called alongside 
        --binary-ped-prefix.
    **--bim** *<bim_filename>*
        Argument used to define the filename of the plink bim file (.bim). Must be 
        called alongside --binary-ped and --fam. Cannot be called alongside 
        --binary-ped-prefix.
    
    #############################
    Output Command-line Arguments
    #############################
    **--out-format** *<output_format>*
        Argument used to define the output format. Supported formats include: gzip
        compressed (gzipped); standard uncompresed (standard); single-precision binary
        (bin32); and double-precision binary (bin64). Please note that both binary formats
        are only supported when called with the square **--lf-format**. By default gzip
        compressed files are produced.
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
        Cannot be used if multiple output files are created.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--overwrite**
        Argument used to define if previous output should be overwritten.
    
    ###############################
    Basic LD Command-line Arguments
    ###############################
    **--ld-statistic** *<r, r2>*
        Argument used to define the correlation statistic to report. Two options are 
        supported: the raw inter-variant allele count correlations (r) and squared 
        correlations (r2).
    **--ld-format** *<square, square-zero, triangle, table>*
        Argument used to define the matrix result format. Four formats are supported:
        A symmetric matrix (square); a square matrix in which the cells of the upper 
        right triangle are zeroed out (square-zero); only the lower-triangular of the 
        matrix (triangle); and the matrix as a table (table). 
	**--ld-window-snps** *<snp_int>*
		Argument used to define the maximum number of SNPs between LD comparisons.
	**--ld-window-kb** *<snp_int>*
		Argument used to define the maximum distance in bp between LD comparisons.
	**--ld-window-cm** *<snp_int>*
		Argument used to define the maximum distance in cM between LD comparisons.
   
    ****************************
    Table Command-line Arguments
    ****************************
	Please note that the following arguments may only be used with **--ld-format** table.

	**--table-d-statistic** *<dprime, dprime-signed, d>*
        Argument used to add the specified D statistic to table-formatted results. Three options
        are supported: the absolute value of Lewontin's D-prime statistic (dprime); Lewontin's 
        D-prime statistic (dprime-signed); and the value of D prior to division by D\ :sub:`max` 
        (d).  
	**--table-in-phase**
		Argument used to add in-phase allele pairs to table-formatted results.
	**--table-maf**
		Argument used to add MAF values to table-formatted results.
	**--table-r2-threshold** *<r2_float>*
		Argument used to define the threshold for filtering pairs of r2 values.
	**--table-snp** *<snp_str>* *<snp1_str, snp2_str, etc.>*
		Argument used to define one or more SNP(s) for LD analysis. This argument may be used 
        multiple times if desired.
    **--table-snps** *<snp_filename>*
		Argument used to define a file with one or more SNP(s) for LD analysis.
    
    ##########################
    Example Command-line Usage
    ##########################
    Command-line to calculate Lewontin's D-prime statistic

    .. code-block:: bash
        
        python plink_ld.py --ped-prefix hapmap1 --ld-format table --ld-statistic r2 --table-d-statistic dprime
'''

import os
import sys
import subprocess
import shutil
import argparse
import glob
import copy
import logging

# Call PPP-based scripts
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir,'pppipe')))

from pgpipe.logging_module import initLogger, logArgs
from pgpipe.plink import *

def plink_argument_parser(passed_arguments):
    '''Phase Argument Parser - Assigns arguments for vcftools from command line.
    Depending on the argument in question, a default value may be specified'''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    plink_parser = argparse.ArgumentParser()

    plink_prefix = plink_parser.add_mutually_exclusive_group()

    # Input PED arguments
    plink_parser.add_argument("--ped", dest = 'ped_filename', help = "Input PED filename", type = str, action = parser_confirm_file())
    plink_parser.add_argument("--map", dest = 'map_filename', help = "Input MAP filename. Called alongside --ped", type = str, action = parser_confirm_file())
    plink_prefix.add_argument("--ped-prefix", help = "Input PED filename prefix", type = str)

    # Input BED arguments
    plink_parser.add_argument("--binary-ped", dest = 'bed_filename', help = "Input Binary-PED (i.e. BED) filename", type = str, action = parser_confirm_file())
    plink_parser.add_argument("--fam", dest = 'fam_filename', help = "Input FAM filename. Called alongside --bed", type = str, action = parser_confirm_file())
    plink_parser.add_argument("--bim", dest = 'bim_filename', help = "Input BIM filename. Called alongside --bed", type = str, action = parser_confirm_file())
    plink_prefix.add_argument("--binary-ped-prefix", dest = 'bed_prefix', help = "Input Binary-PED (i.e. BED) filename prefix", type = str)

    # Method arguments
    ld_statistics = ['r', 'r2']
    ld_default = 'r2'
    plink_parser.add_argument('--ld-statistic', metavar = metavar_list(ld_statistics), help = 'Specifies the LD statistic', type = str, choices = ld_statistics, default = ld_default)

    shape_formats = ['triangle', 'square', 'square-zero', 'table']
    shape_default = 'table'
    plink_parser.add_argument('--ld-format', metavar = metavar_list(shape_formats), help = 'Specifies the format of the LD results', type = str, choices = shape_formats, default = shape_default)

    # LD-based arguments
    plink_parser.add_argument('--ld-window-snps', help = 'Sets the maximum number of SNPs between LD comparisons', type = int)
    plink_parser.add_argument('--ld-window-kb', help = 'Sets the maximum distance in bp between LD comparisons', type = int)
    plink_parser.add_argument('--ld-window-cm', help = 'Sets the maximum distance in cM between LD comparisons', type = float)

    # Table specific options
    d_statistics = ['d', 'dprime', 'dprime-signed']
    plink_parser.add_argument('--table-d-statistic', metavar = metavar_list(d_statistics), help = 'Adds the specified D statistic to table-formatted results', type = str, choices = d_statistics)

    plink_parser.add_argument('--table-in-phase', help = 'Adds in-phase allele pairs to table-formatted results', action = 'store_true')
    plink_parser.add_argument('--table-maf', help = 'Adds MAF to table-formatted results', action = 'store_true')
    plink_parser.add_argument('--table-r2-threshold', help = "Sets the threshold for filtering pairs of r2 values", type = float)
    plink_parser.add_argument('--table-snp', help = "Specifies one or more SNP(s) for LD analysis", type = str, action = parser_add_to_list())
    plink_parser.add_argument('--table-snps', help = "Specifies a file with one or more SNP(s) for LD analysis", type = str, action = parser_confirm_file())

    # Output arguments
    output_formats = ['standard', 'gzipped', 'bin64', 'bin32']
    output_default = 'gzipped'
    plink_parser.add_argument('--out-format', metavar = metavar_list(output_formats), help = 'Specifies the output format.', type = str, choices = output_formats, default = output_default)

    plink_parser.add_argument('--out', help = 'Defines the output filename. Only usable with vcf-based output')
    plink_parser.add_argument('--out-prefix', help = 'Defines the output filename prefix', default = 'out')

    # General arguments.
    plink_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')
    plink_parser.add_argument('--threads', help = "Set the number of threads", type = int, default = 1)

    if passed_arguments:
        return plink_parser.parse_args(passed_arguments)
    else:
        if len(sys.argv) == 1:
            plink_parser.print_help(sys.stderr)
            sys.exit(1)
        return plink_parser.parse_args()

def run (passed_arguments = []):
    '''
        PLINK-based LD Analysis

        Automates plink-based LD analysis. The function has arguments for
        selecting the LD statistic to be used, the output format of the
        analysis, in addition to many of the optional arguments from plink.

        Parameters
        ----------
        --vcf : str
            Specifies the input VCF filename
        --vcf-fid : str
            Specifies the family ID for all samples
        --ped : str
            Specifies the input PED filename
        --map : str
            Specifies the input MAP filename. Called alongside --ped
        --ped-prefix : str
            Specifies the input PED filename prefix
        --bed : str
            Specifies the input BED filename
        --map : str
            Specifies the input FAM filename. Called alongside --bed
        --bim : str
            Specifies the input BIM filename. Called alongside --bed
        --bed-prefix : str
            Specifies the input BED filename prefix
        --ld-statistic : str
            Specifies the LD statistic {r, r2} (Default: r2)
        --lf-format : str
            Specifies the format of the LD results {triangle, square,
            square-zero, table} (Default: table)
        --ld-window-snps : int
            Sets the maximum number of SNPs between LD comparisons
        --ld-window-kb : int
            Sets the maximum distance in bp between LD comparisons
        --ld-window-cm : float
            Sets the maximum distance in cM between LD comparisons
        --table-d-statistic : str
            Adds the specified D statistic to table-formatted results {d,
            dprime, dprime-signed}
        --table-in-phase
            Adds in-phase allele pairs to table-formatted results
        --table-maf
            Adds MAF to table-formatted results
        --table-r2-threshold : float
            Sets the threshold for filtering pairs of r2 values"
        --table-snp : str, list
            Specifies one or more SNP(s) for LD analysis
        --table-snps : str
            Specifies a file with one or more SNP(s) for LD analysis
        --out-format : str
            Specifies the output format {standard, gzipped, bin64, bin32}
            (Default: gzipped)
        --out : str
            Specifies the output filename. Only usable with vcf-based output
        --out-prefix : str
            Specifies the output filename prefix
        --overwrite
            Overwrite previous output files
        --threads : int
            Set the number of threads

        Returns
        -------
        output : file
            Statistic file output
        log : file
            Log file output

        Raises
        ------
        IOError
            Input file does not exist

    '''

    # Grab plink arguments from command line
    plink_args = plink_argument_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(plink_args, func_name = 'plink_ld')

    # List to hold general plink arguments
    plink_general_args = []

    # Assign the number of threads
    plink_general_args.extend(['--threads', plink_args.threads])

    # List to hold the input argument(s)
    plink_input_args = []

    # Check if the user has selected the table ld format
    if plink_args.ld_format == 'table':

        # Check if the output format isn't supported
        if plink_args.out_format not in ['standard', 'gzipped']:
            raise Exception('The %s output only supports the square ld format (i.e. --lf-format square)' % plink_args.out_format)

        # Check if a r2 threshold was specified alonside the r2 statistic
        if plink_args.table_r2_threshold and plink_args.ld_statistic != 'r2':
            raise Exception('The --table-r2-threshold argument is only compatible with the r2 ld statistic')

    # Check if the user has selected an ld format that isn't table
    else:

        # Check if the output format isn't supported
        if plink_args.out_format != 'standard':
            raise Exception('The %s ld format is incompatible with the standard output format' % plink_args.ld_format)

        # Check if a d statistic was specified, which is incompatible
        if plink_args.table_d_statistic:
            raise Exception('The --table-d-statistic argument is incompatible with the %s ld format' % plink_args.ld_format)

        # Check if in-phase allele pairs were specified, which are incompatible
        if plink_args.table_in_phase:
            raise Exception('The --table-in-phase argument is incompatible with the %s ld format' % plink_args.ld_format)

        # Check if in-phase allele pairs were specified, which are incompatible
        if plink_args.table_maf:
            raise Exception('The --table-maf argument is incompatible with the %s ld format' % plink_args.ld_format)

        # Check if a r2 threshold was specified, which is incompatible
        if plink_args.table_r2_threshold:
            raise Exception('The --table-r2-threshold argument is incompatible with the %s ld format' % plink_args.ld_format)

        # Check if a r2 threshold was specified, which is incompatible
        if plink_args.table_snp:
            raise Exception('The --table-snp argument is incompatible with the %s ld format' % plink_args.ld_format)

        # Check if a r2 threshold was specified, which is incompatible
        if plink_args.table_snps:
            raise Exception('The --table-snps argument is incompatible with the %s ld format' % plink_args.ld_format)

    # Confirm if a prefix has been specified and the files exist
    if plink_args.ped_prefix and confirm_ped_prefix(plink_args.ped_prefix):

        # Assign bed input arguments from a prefix
        plink_input_args = ['--file', plink_args.ped_prefix]

    # Confirm if a prefix has been specified and the files exist
    elif plink_args.bed_prefix and confirm_ped_prefix(plink_args.bed_prefix):

        # Assign the bed input from a prefix
        plink_input_args = ['--bfile', plink_args.bed_prefix]

    # Confirm if required ped files exist
    elif plink_args.ped_filename and confirm_ped_files(plink_args.ped_filename, plink_args.map_filename):

        # Assign the ped input
        plink_input_args = ['--ped', plink_args.ped_filename, '--map', plink_args.map_filename]

    # Confirm if required binary-ped files exist
    elif plink_args.bed_filename and confirm_bed_files(plink_args.bed_filename, plink_args.bim_filename, plink_args.fam_filename):

        # Assign bed input
        plink_input_args = ['--bed', plink_args.bed_filename, '--bim', plink_args.bim_filename, '--fam', plink_args.fam_filename]

    else:
        raise Exception('No PED or Binary-PED input assigned')


    logging.info('Input parameters assigned')

    # List to hold the ld argument(s)
    plink_ld_args = []

    # Add the ld statistic method to the ld argument list
    plink_ld_args.append('--%s' % plink_args.ld_statistic)

    # Check if the format is either triangle or square
    if plink_args.ld_format in ['triangle', 'square']:

        # Add the ld format to the ld argument list
        plink_ld_args.append(plink_args.ld_format)

    # Check if the format is square-zero
    elif plink_args.ld_format == 'square-zero':

        # Add the ld format to the ld argument list
        plink_ld_args.append('square0')

    # Check if the format is a table
    elif plink_args.ld_format == 'table':

        # Add the ld format to the ld argument list
        plink_ld_args.append('inter-chr')

    # Check if an unsupported format was found and raise an exception
    else:
        raise Exception('The format specified (%s) is not supported' % plink_args.ld_format)

    # String to hold output suffix
    plink_output_suffix = 'ld'

    # Check if the output format is gzipped
    if plink_args.out_format == 'gzipped':

        # Add the output format to the ld argument list
        plink_ld_args.append('gz')

        # Add gz to the output suffix
        plink_output_suffix += '.gz'

    # Check if the output format is double-percision binary
    elif plink_args.out_format == 'bed64':

        # Add the output format to the ld argument list
        plink_ld_args.append('bed')

        # Add bed to the output suffix
        plink_output_suffix += '.bed'

    # Check if the output format is single-percision binary
    elif plink_args.out_format == 'bed32':

        # Add the output format to the ld argument list
        plink_ld_args.append('bed4')

        # Add bed to the output suffix
        plink_output_suffix += '.bed'

    elif plink_args.out_format == 'standard':

        # No output format to add to the ld argument list, but not error
        pass

    # Check if an unsupported format was found and raise an exception
    else:
        raise Exception('The format specified (%s) is not supported' % plink_args.out_format)

    # Check if the user has specified the table-specific in-phase allele pairs argument
    if plink_args.table_in_phase:

        # Add the table-specific in-phase allele pairs to the ld argument list
        plink_ld_args.append('in-phase')

    # Check if the user has specified the table-specific d statistic argument
    if plink_args.table_d_statistic:

        # Add the table-specific d statistic to the ld argument list
        plink_ld_args.append(plink_args.table_d_statistic)

    # Check if the user has specified the table-specific maf argument
    if plink_args.table_maf:

        # Add the table-specific maf to the ld argument list
        plink_ld_args.append('with-freqs')

    # Check if the user set a maximum number of SNPs between LD comparisons
    if plink_args.ld_window_snps:

        # Assign the maximum SNP arguments
        plink_ld_args.extend(['--ld-window', plink_args.ld_window_snps])

    # Check if the user set a maximum distance in bp between LD comparisons
    if plink_args.ld_window_kb:

        # Assign the maximum bp distance arguments
        plink_ld_args.extend(['--ld-window-kb', plink_args.ld_window_kb])

    # Check if the user set a maximum distance in cM between LD comparisons
    if plink_args.ld_window_cm:

        # Assign the maximum cm distance arguments
        plink_ld_args.extend(['--ld-window-cm', plink_args.ld_window_cm])

    # Check if the user set a r2 threshold
    if plink_args.table_r2_threshold:

        # Assign the r2 threshold
        plink_ld_args.extend(['--ld-window-r2', plink_args.table_r2_threshold])


    # Check if the user has specified SNP(s) to analyze
    if plink_args.table_snp:

        # Check if there is only a single snp
        if len(plink_args.table_snp) == 1 and '-' not in plink_args.table_snp[0]:

            # Assign the snp
            plink_ld_args.extend(['--ld-snp', plink_args.table_snp[0]])

        # Check if there is more than a single snp
        else:

            # Assign the snps
            plink_ld_args.extend(['--ld-snps', ', '.join(plink_args.table_snp)])

    # Check if the user has specified a SNP file
    if plink_args.table_snps:

        # Assign the snp file
        plink_ld_args.extend(['--ld-snp-list', plink_args.table_snps])

    logging.info('LD parameters assigned')

    # Check if the has specified the output filename, without a prefix
    if plink_args.out and '--out-prefix' not in passed_arguments and '--out-prefix' not in sys.argv:

        # Assign a prefix based on the output filename
        phase_args.out_prefix = copy.deepcopy(phase_args.out)

    # Assign the prefix argument
    plink_output_args = ['--out', plink_args.out_prefix]

    # Assign the expected output filename
    plink_expected_output = '%s.%s' % (plink_args.out_prefix, plink_output_suffix)

    # Check if previous output should not be overwritten
    if not plink_args.overwrite:

        # Check if previous output exists
        if os.path.exists(plink_expected_output):
            raise Exception('%s already exists. Add --overwrite to ignore' % plink_expected_output)

    logging.info('Output parameters assigned')

    # Call plink with the selected arguments
    call_plink(plink_input_args + plink_ld_args + plink_output_args + plink_general_args)

    # Rename output to plink_args.out, if specified
    if plink_args.out:
        shutil.move(plink_expected_output, plink_args.out)
        shutil.move(plink_args.out_prefix + '.log', plink_args.out + '.log')

    # Rename log using plink_args.out_prefix
    else:
        shutil.move(plink_args.out_prefix + '.log', '%s.log' % plink_expected_output)


if __name__ == "__main__":
    initLogger()
    run()



from __future__ import print_function
import sys
import pysam
from math import ceil,floor
from pgpipe.vcf_reader_func import VcfReader
from pgpipe.genome_region import Region, RegionList
import logging

def findPrevLastSnps(vcfr,region,window=1000):
    left_out = 0
    i = region.start-1
    r = Region(i-window,i,region.chrom)
    tl = vcfr.getRecordList(r)
    if len(tl) != 0:
        left_out = tl[-1].pos
    else:
        left_out = i - window
    right_out = 0
    i = region.end+1
    r = Region(i,i+window,region.chrom)
    tl = vcfr.getRecordList(r)
    if len(tl) != 0:
        right_out = tl[0].pos
    else:
        right_out = i + window
    return left_out,right_out

#Given a VCF file and a region file, extend regions to
#length halfway between border SNPs inside/outside of the range

vcf_name = str(sys.argv[1])
lil_name = str(sys.argv[2])

vcfr = VcfReader(vcf_name)
lilr = VcfReader(lil_name)

lilrecs = lilr.getRecordList()


region = Region(lilrecs[0].pos-1,lilrecs[-1].pos+1,lilrecs[0].chrom)

rl = vcfr.getRecordList(region)
if len(rl) == 0:
    logging.warning("Region at %s has no variants" % region.toStr())
    print (region.toStr(sep='\t'))
    exit()
left_in = rl[0].pos
right_in = rl[-1].pos
left_out,right_out = findPrevLastSnps(vcfr,region)
left_half = int(ceil((left_in+left_out)/2))
right_half = int(floor((right_in+right_out)/2))
print (region.chrom,left_half,right_half,left_out,left_in,right_in,right_out,sep='\t')



import os
import sys
import json
import argparse
import logging
import itertools
from pgpipe.model import Model, ModelFile, write_model_file

from collections import defaultdict
from collections import OrderedDict

# Insert Jared's directory path, required for calling Jared's functions. Change when directory structure changes.
#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

from pgpipe.logging_module import initLogger

def model_creator_parser (passed_arguments):
    '''VCF Argument Parser - Assigns arguments from command line'''

    def parser_dict_list_append ():
        '''Custom action to house data as defaultdict list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                if getattr(args, self.dest):
                    if values[1] not in getattr(args, self.dest)[values[0]]:
                        # Append the argument with the file
                        getattr(args, self.dest)[values[0]].append(values[1])
                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(list)
                    arg_dict[values[0]].append(values[1])
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_list_extend ():
        '''Custom action to house data as defaultdict list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                if getattr(args, self.dest):
                    for value in values[1:]:
                        if value not in getattr(args, self.dest)[values[0]]:
                            # Append the argument with the file
                            getattr(args, self.dest)[values[0]].append(value)
                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(list)
                    arg_dict[values[0]].extend(values[1:])
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_str ():
        '''Custom action to house data as defaultdict int'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):
                # Assign the passed value
                if getattr(args, self.dest):
                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]] = values[1]

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(str)
                    arg_dict[values[0]] = values[1]
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_int ():
        '''Custom action to house data as defaultdict int'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):

                # Check the passed value is an int
                try:
                    int(values[1])
                except:
                    raise Exception('--%s only accepts integers' % self.dest)

                # Assign the passed value
                if getattr(args, self.dest):
                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]] = int(values[1])

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(int)
                    arg_dict[values[0]] = int(values[1])
                    setattr(args, self.dest, arg_dict)
        return customAction

    def parser_dict_file ():
        '''Custom action to house data as defaultdict int'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, values, option_string=None):

                if not os.path.isfile(values[1]):
                    raise IOError('%s not found' % values[1])

                # Assign the passed value
                if getattr(args, self.dest):
                    # Append the argument with the file
                    getattr(args, self.dest)[values[0]] = values[1]

                else:
                    # Set the argument with the file (as list)
                    arg_dict = defaultdict(str)
                    arg_dict[values[0]] = values[1]
                    setattr(args, self.dest, arg_dict)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    model_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # General model arguments
    model_parser.add_argument('--model', help = 'Defines the name of a model', type = str, action = 'append', required = True)

    # Tree arguments
    model_parser.add_argument('--model-tree', help = 'Assign population tree string to a model, in Newick format', type = str, nargs = 2, action = parser_dict_str())
    model_parser.add_argument('--model-tree-file', help = 'Assign population tree file to a model, in Newick format', type = str, nargs = 2, action = parser_dict_file())

    # Population arguments
    model_parser.add_argument('--model-pop', dest = 'pops', help = 'Assign a population name to a model', type = str, nargs = 2, action = parser_dict_list_append())
    model_parser.add_argument('--model-pops', dest = 'pops', help = 'Assign multiple population names to a model', type = str, nargs = '+', action = parser_dict_list_extend())
    model_parser.add_argument('--model-pop-file', help = 'Assign multiple population names to a model usign a file', type = str, nargs = 2, action = parser_dict_file())

    # Individual arguments
    model_parser.add_argument('--pop-ind', dest = 'inds', help = 'Assign an individual name to a population', type = str, nargs = 2, action = parser_dict_list_append())
    model_parser.add_argument('--pop-inds', dest = 'inds', help = 'Assign multiple individual names to a population', type = str, nargs = '+', action = parser_dict_list_extend())
    model_parser.add_argument('--pop-ind-file', help = 'Assign multiple individual names to a population using a file', type = str, nargs = 2, action = parser_dict_file())

    # Output Arguments
    model_parser.add_argument('--out', help = 'Specifies the complete output filename.', type = str, default = 'out.model')
    model_parser.add_argument('--overwrite', help = "Specifies if previous output files should be overwritten", action = 'store_true')

    if passed_arguments:
        return model_parser.parse_args(passed_arguments)
    else:
        return model_parser.parse_args()

def logArgs(args, pipeline_function):
    '''Logs arguments from argparse system. May be replaced with a general function in logging_module'''
    logging.info('Arguments for %s:' % pipeline_function)
    for k in vars(args):
        logging.info('Arguments %s: %s' % (k, vars(args)[k]))

def incompatible_duplicates_check (term, *arguments_to_test):

    # Duplicate count
    duplicates_count = 0

    # Loop the passed arguments
    for argument_to_test in arguments_to_test:

        # Check if the argument has been populated
        if argument_to_test:

            # Check if the argument has the term
            if term in argument_to_test:

                # Add to the duplicate count if the incompatible term is found
                duplicates_count += 1

    # Check if there were duplicates_count
    if duplicates_count > 1:
        return True
    else:
        return False

def assigment_check (term, *arguments_to_be_assigned):

    # Term assignment boolean
    term_assigned = False

    # Loop the passed arguments
    for argument_to_be_assigned in arguments_to_be_assigned:

        # Check if the argument has been populated
        if argument_to_be_assigned:

            # Check if the argument has the term
            if term in argument_to_be_assigned:

                # Mark the term assigned
                term_assigned = True

    # Return the boolean
    return term_assigned

def dict_argument_to_list (term, list_argument):

    # Check if either the argument or term was not assigned
    if not list_argument or term not in list_argument:
        return []

    # Return data as list
    return list_argument[term]

def file_dict_argument_to_list (term, file_argument):

    # Check if either the argument or term was not assigned
    if not file_argument or term not in file_argument:
        return []

    # List to hold the file data
    data_list = []

    # Check if running via python 3
    if sys.version_info[0] == 3:

        # Open the file and save the data
        with open(file_argument[term], 'r') as data_file:
            for data_line in data_file:
                data_list.append(data_line.strip())

    # Check if running via python 2
    else:

        # Open the file and save the data
        with open(file_argument[term], 'rb') as data_file:
            for data_line in data_file:
                data_list.append(data_line.strip())

    # Return data as list
    return data_list

def run (passed_arguments = []):
    '''

    Parameters
    ----------
    --model : str
        Name of the model
    --model-tree : str, str
        Assign newick-formatted population tree to model
        Example: --model-tree model_name newick_tree
    --model-tree-file : str, str
        Assign newick-formatted population tree file to model
        Example: --model-tree-file model_name newick_file
    --model-pop : str, str
        Assign single population name to a model
        Example: --model-pop model_name pop_name
    --model-pops : str, list
        Assign multiple population names to a model
        Example: --model-pops model_name pop_name1 pop_name2 etc.
    --model-pop-file : str, str
        Assign population names to a model using a file
        Example: --model-pop-file model_name pop_file
    --pop-ind : str, str
        Assign single individual name to a population
        Example: --pop-ind pop_name ind_name
    --pop-inds : str, list
        Assign multiple population names to a model
        Example: --pop-inds pop_name ind_name1 ind_name2 etc.
    --pop-ind-file : str, str
        Assign population names to a model using a file
        Example: --pop-ind-file pop_name ind_file
    --out : str
        Filename of the model output

    Returns
    -------

    Raises
    ------
    Exception
        No tree assigned to model
    Exception
        Multiple trees assigned to model
    Exception
        No population assigned to model
    Exception
        No individuals assigned to population

    '''

    # Grab VCF arguments from command line
    creator_args = model_creator_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(creator_args, 'model_creator')

    # Create the model file
    model_file = ModelFile()

    # Loop each model specified to confirm parameters are valid
    for model in creator_args.model:

        # Check that a tree has been assigned to the model
        if not assigment_check(model, creator_args.model_tree_file, creator_args.model_tree):
            raise Exception('No tree assigned to %s' % model)

        # Check that multiple trees have not been assigned to the model
        if incompatible_duplicates_check(model, creator_args.model_tree_file, creator_args.model_tree):
            raise Exception('Multiple trees assigned to %s' % model)

        # Check if a pop assignment arguments has been used
        if not assigment_check(model, creator_args.pops, creator_args.model_pop_file):
            raise Exception('No populations assigned to %s' % model)

        # Create the model
        model_data = Model(name = model)

        logging.info('Model (%s) created' % model)

        # Convert the inds dict into a list
        pop_list = dict_argument_to_list(model, creator_args.pops)

        # Convert the pop file into a list
        pop_file_list = file_dict_argument_to_list(model, creator_args.model_pop_file)

        # Combine the pop arguments
        combined_pops = list(set(pop_list + pop_file_list))

        logging.info('Populations assigned for model (%s)' % model)

        # Assign the tree
        model_data.assign_tree(tree = '')

        # Loop each population in the model
        for pop in combined_pops:

            # Check if a individual assignment arguments has been used
            if not assigment_check(pop, creator_args.inds, creator_args.pop_ind_file):
                raise Exception('No individuals assigned to %s' % pop)

            # Convert the inds dict into a list
            ind_list = dict_argument_to_list(pop, creator_args.inds)

            # Convert the inds file into a list
            ind_file_list = file_dict_argument_to_list(pop, creator_args.pop_ind_file)

            # Combine the pop arguments
            combined_inds = list(set(ind_list + ind_file_list))

            logging.info('Individuals assigned for population (%s)' % pop)

            # Assign the population
            model_data.assign_pop(pop = pop, inds = combined_inds)

            logging.info('Model (%s) assigned to model file' % model)

        # Add the model to the model file
        model_file[model] = model_data

    # Create the model file
    write_model_file(model_file, creator_args.out, overwrite = creator_args.overwrite)


if __name__ == "__main__":
    initLogger()
    run()



import pandas as pd
import matplotlib
matplotlib.use('agg')
import subprocess
from io import StringIO
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

# Color blind color cycle hex codes
CB_color_cycle = ['#ffb000', '#dc267f', '#648fff', '#fe6100', '#785ef0']


def manhattan_plot(in_file):
    try:
        data = pd.read_table(in_file, sep="\t")
        data = data.dropna()
        fig = plt.figure()
        ax = fig.add_subplot(111)
        data.plot(kind='scatter', x='POS', y='WEIR_AND_COCKERHAM_FST', ax=ax)
        ax.set_xlabel('Position')
        fig.savefig(in_file + ".pdf", bbox_inches='tight')
    except FileNotFoundError as e:
        print(e+' not found')


def admixture_barplot(in_file, numinds):
    matrix = []
    try:
        with open(in_file, 'r') as file:
            for line in file:
                if 'Miss' in line:
                    for i in range(numinds):
                        row = (next(file).strip().replace(':', '').split())
                        matrix.append(row)
        data = pd.DataFrame(matrix, columns=['Index', 'Label', '(%Miss)', 'Pop', 'A', 'B'])
        data = data.drop('Index',  axis=1)

        fig = plt.figure()
        r = data.index.tolist()

        # convert series A and B dataset to list
        series_a = data[data.columns[3]].apply(pd.to_numeric).tolist()
        series_b = data[data.columns[4]].apply(pd.to_numeric).tolist()
        # Create bar plot for series A and B data
        plt.bar(r, series_a,  width=1.0, color='#b5ffb9')
        plt.bar(r, series_b, bottom=series_a, width=1.0, color='#f9bc86')
        # Add label
        plt.xlabel("Label")
        plt.ylabel("Frequency")
        # Save plot as pdf file
        fig.savefig(in_file + ".pdf", bbox_inches='tight')
    except FileNotFoundError as e:
        print(e)


def pca_plot(in_file):
    try:
        df = pd.read_table(in_file, index_col=0, header=0)

        features = list(df.columns.values)
        # Separating out the features
        data = df.loc[:, features].values

        # Standardizing the features
        data = StandardScaler().fit_transform(data)
        pca = PCA(n_components=2)
        prncpl_comp = pca.fit_transform(data)
        prncpl_df = pd.DataFrame(data=prncpl_comp, columns=['PC1', 'PC2'])
        fig = plt.figure()
        plt.xlabel('Principal Component 1')
        plt.ylabel('Principal Component 2')
        plt.title('2 component PCA', fontsize=20)

        plt.scatter(prncpl_df['PC1'], prncpl_df['PC2'])
        plt.legend('Samples')
        plt.grid()
        fig.savefig(in_file + ".pdf", bbox_inches='tight')
    except FileNotFoundError as e:
        print(e)


# Function for plotting graphs for admixture output data set
def bar_plot(in_file):
    try:
        a = in_file
        df = pd.read_table(a, header=None, delim_whitespace=True)
        df.plot.bar(stacked=True, figsize=(23, 5), color=CB_color_cycle)
        plt.savefig(in_file + ".pdf", bbox_inches='tight')
    except FileNotFoundError as e:
        print(e)


# Function for plotting graphs for ima2p output data sets
def do_plots(in_file):
    #try:
        output1 = subprocess.run("grep -B 1000 -m 1 \"SumP\" " + in_file + " | grep -v \"SumP\"", shell=True,
                                 stdout=subprocess.PIPE, universal_newlines=True)
        output2 = subprocess.run("grep -B 1000 -m 2 \"SumP\" " + in_file + " | grep -v \"SumP\" | tail -n1000",
                                 shell=True, stdout=subprocess.PIPE, universal_newlines=True)

        t0 = pd.read_table(StringIO(output1.stdout), sep="\t", header=None)
        t0.drop(t0.columns[[0]], axis=1, inplace=True)

        q0 = pd.read_table(StringIO(output2.stdout), sep="\t", header=None)
        q0.drop(q0.columns[[0]], axis=1, inplace=True)

        fig, ax = plt.subplots(nrows=3, ncols=2)

        ax[0, 0].plot(t0[2], t0[3])
        ax[0, 0].set_xlabel("Divergence times (t0)")
        ax[0, 0].set_ylabel("Frequency")

        ax[0, 1].plot(q0[2], q0[3])
        ax[0, 1].set_xlabel("Population size (q0)")
        ax[0, 1].set_ylabel("Frequency")

        ax[1, 0].plot(q0[4], q0[5])
        ax[1, 0].set_xlabel("Population size (q1)")
        ax[1, 0].set_ylabel("Frequency")

        ax[1, 1].plot(q0[6], q0[7])
        ax[1, 1].set_xlabel("Population size (q2)")
        ax[1, 1].set_ylabel("Frequency")

        ax[2, 0].plot(q0[8], q0[9])
        ax[2, 0].set_xlabel("Migration rate (m0->1)")
        ax[2, 0].set_ylabel("Frequency")

        ax[2, 1].plot(q0[10], q0[11])
        ax[2, 1].set_xlabel("Migration rate (m1->0)")
        ax[2, 1].set_ylabel("Frequency")

        plt.tight_layout()
        plt.savefig(in_file + ".pdf")
    #except:
    #    print(in_file + ": No such file or directory")



#!/usr/bin/env python
'''
    Automates various simple file conversions. Currently the function is 
    capable of converting between VCF-based formats (i.e. VCF, compressed-VCF, 
    and BCF) and PLINK-based formats (i.e. PED and Binary-PED). Additional 
    formats will be added as needed. 

    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<vcf_filename>*
        Argument used to define the filename of the VCF file.
    **--vcf-fid** *<fid_str>*
        Argument used to define the family ID for all VCF samples.
    **--ped-prefix** *<ped_prefix>*
        Argument used to define the filename prefix of both PED and MAP files.
    **--ped** *<ped_filename>*
        Argument used to define the filename of the PED file. Called alongside
        **--map**.
    **--map** *<map_filename>*
        Argument used to define the filename of the MAP file. Called alongside
        **--ped**.
    **--binary-ped-prefix** *<binary_ped_prefix>*
        Argument used to define the filename prefix of the Binary-PED, FAM,
        and BIM files.
    **--binary-ped** *<ped_filename>*
        Argument used to define the filename of the Binary-PED (i.e. BED) file. 
        Called alongside **--fam** and **--bim**.
    **--fam** *<fam_filename>*
        Argument used to define the filename of the FAM file. Called alongside
        **--binary-ped** and **--bim**.
    **--bim** *<bim_filename>*
        Argument used to define the filename of the BIM file. Called alongside
        **--binary-ped** and **--fam**.
   
    #############################
    Output Command-line Arguments
    #############################
    **--out** *<output_filename>*
        Argument used to define the complete output filename, overrides **--out-prefix**.
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension).
    **--out-format** *<vcf, vcf.gz, bcf, ped, ped-12, binary-ped, eigenstrat>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz); BCF (bcf); PLINK text file (ped); PLINK "12" 
        coded text file (ped-12); binary PLINK file (binary-ped); and eigenstrat file
        (eigenstrat).
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    ############################
    Other Command-line Arguments
    ############################
    **--delete-original**
        Argument used to define that the original file should be deleted once converted.
    **--threads** *<thread_int>*
        Argument used to define the number of threads. This argument is currently only 
        supported by conversions to/from PED and Binary-PED.
'''

import os
import sys
import subprocess
import shutil
import argparse
import glob
import logging

#sys.path.insert(0, os.path.abspath(os.path.join(os.pardir, 'pppipe')))

# Import PPP modules and scripts
from pgpipe.admixtools import *
from pgpipe.bcftools import convert_vcf, log_bcftools_reference
from pgpipe.plink import convert_ped, convert_bed, convert_vcf_to_plink, log_plink_reference
from pgpipe.vcf_reader_func import checkFormat
from pgpipe.logging_module import initLogger, logArgs


def convert_argument_parser(passed_arguments):
    '''
    Convert Argument Parser

    Assign the parameters for the convert function using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    # Create the parser
    convert_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input VCF argument
    convert_parser.add_argument("--vcf", help = "Defines the filename of the VCF file", type = str, action = parser_confirm_file())

    # Sets a family ID for the samples in the VCF
    convert_parser.add_argument("--vcf-fid", dest = 'vcf_fid', help = "Defines the family ID for all VCF samples", type = str)

    # Input FASTA argument
    #convert_parser.add_argument("--fasta", help = "Input FASTA filename", type = str, action = parser_confirm_file())

    # Input IM argument
    #convert_parser.add_argument("--im", help = "Input IM filename", type = str, action = parser_confirm_file())

    # Input PED arguments
    convert_parser.add_argument("--ped", dest = 'ped_filename', help = "Defines the filename of the PED file. Called alongside --map", type = str, action = parser_confirm_file())
    convert_parser.add_argument("--map", dest = 'map_filename', help = "Defines the filename of the MAP file. Called alongside --ped", type = str, action = parser_confirm_file())

    # Input BED arguments
    convert_parser.add_argument("--binary-ped", dest = 'bed_filename', help = "Defines the filename of the Binary-PED (i.e. BED) file. Called alongside --fam and --bim", type = str, action = parser_confirm_file())
    convert_parser.add_argument("--fam", dest = 'fam_filename', help = "Defines the filename of the FAM file. Called alongside --binary-ped and --bim", type = str, action = parser_confirm_file())
    convert_parser.add_argument("--bim", dest = 'bim_filename', help = "Defines the filename of the BIM file. Called alongside --binary-ped and --fam", type = str, action = parser_confirm_file())

    # Input PED/BED prefix arguments
    convert_prefix = convert_parser.add_mutually_exclusive_group()
    convert_prefix.add_argument("--ped-prefix", help = "Defines the filename prefix of both PED and MAP files", type = str)
    convert_prefix.add_argument("--binary-ped-prefix", dest = 'bed_prefix', help = "Defines the filename prefix of the Binary-PED, FAM, and BIM files", type = str)

    # Output arguments.
    output_formats = ['vcf', 'vcf.gz', 'bcf', 'ped', 'ped-12', 'binary-ped', 'eigenstrat']
    convert_parser.add_argument('--out-format', metavar = metavar_list(output_formats), help = 'Defines the desired output format', type = str, choices = output_formats, required = True)
    convert_parser.add_argument('--out', help = 'Defines the complete output filename, overrides --out-prefix')
    convert_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', default = 'out')

    # Other basic arguments. Expand as needed
    convert_parser.add_argument('--delete-original', help = 'Defines that the original file should be deleted once converted', action='store_false')
    convert_parser.add_argument('--overwrite', help = "Defines if previous output should be overwritten", action = 'store_true')
    convert_parser.add_argument('--threads', help = "Set the number of threads. Only supported with ped", type = int, default = 1)

    if passed_arguments:
        return convert_parser.parse_args(passed_arguments)
    else:
        return convert_parser.parse_args()

def check_if_conversion (input_format, out_format):
    '''
        Checks that the input and output formats differ

        Parameters
        ----------
        input_format : str
            File format of the input

        out_format : str
            File format of the output

        Raises
        ------
        Exception
            If the input and output formats do not differ
    '''

    # Check if the input and output are the same formats
    if input_format == out_format:
        raise Exception('Input and Output formats are identical. Cannot convert')

def check_conversion_support (out_format, supported_formats):
    '''
        Confirms it is possible to convert the input into the specified output
        format

        Parameters
        ----------
        out_format : str
            File format of the output
        supported_formats : str
            The file formats the input supports (i.e. allows conversion)

        Raises
        ------
        Exception
            If the output format is not supported
    '''

    # Check that output format is supported
    if out_format not in supported_formats:
        raise Exception('Input cannot be converted into the format specified')

def run (passed_arguments = []):
    '''
    File conversion suite

    This function uses the argparse-based function :py:func:`convert_argument_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their respective equivalent. Once all the 
    parameters are assigned, the conversion function of choice is called.

    Parameters
    ----------
    --vcf : str
        Filename of the VCF
    --ped : str
        Filename of the PED
    --map : str
        Filename of the MAP
    --bed : str
        Filename of the Binary-PED
    --map : str
        Filename of the FAM
    --bim : str
        Filename of the BIM
    --ped-prefix : str
        PED file prefix
    --binary-ped-prefix : str
        Binary-PED file prefix
    --out-prefix : str
        Output filename prefix
    --out-format : str
        Output format

    Raises
    ------
    Exception
        Impossible conversion given
    Exception
        Conversion to same format given
    '''

    # Grab plink arguments from command line
    convert_args = convert_argument_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(convert_args, func_name = 'convert')

    logging.info('Conversion parameters assigned')

    # Check if the format is binary-ped, i.e. .bed
    if convert_args.out_format == 'binary-ped':

        # Assign the expected output filename
        expected_out_filename = convert_args.out_prefix + '.bed'

    # Check if the format is ped-12, i.e. .ped
    elif convert_args.out_format == 'ped-12':

        # Assign the expected output filename
        expected_out_filename = convert_args.out_prefix + '.ped'
        
    else:

        # Assign the expected output filename
        expected_out_filename = '%s.%s' % (convert_args.out_prefix, convert_args.out_format)

    # Check if input is specified as a vcf
    if convert_args.vcf:

        # Determine the format of the vcf file (e.g. vcf, vcf.gz, bcf)
        vcf_format = checkFormat(convert_args.vcf)

        # Convert bgzip flag to vcf.gz. Consider changing checkFormat function?
        if vcf_format == 'bgzip':
            
            # Convert to the format used by convert_args.out_format
            vcf_format = 'vcf.gz'

        # Check that the input is in a supported file format
        if vcf_format not in ['vcf', 'vcf.gz', 'bcf']:
            raise Exception('Unknown VCF file format')

        # Check that the input and output formats differ
        check_if_conversion(vcf_format, convert_args.out_format)

        # List of formats that are supported from vcf
        supported_out_format = ['vcf', 'vcf.gz', 'bcf', 'ped', 'ped-12', 'binary-ped']

        # Check that the conversion is supported
        check_conversion_support(convert_args.out_format, supported_out_format)

        logging.info('Convertion method assigned')

        # Check if the output is a vcf
        if convert_args.out_format in ['vcf', 'vcf.gz', 'bcf']:

            convert_vcf(convert_args.vcf, convert_args.out_prefix, convert_args.out_format, convert_args.overwrite, keep_original = True)

            # Add the reference to the log
            log_bcftools_reference(expected_out_filename, append_mode = False)

            # Check if a specific output filename was specified
            if convert_args.out:

                # Rename the vcf file
                shutil.move(expected_out_filename, convert_args.out)

                # Rename the log file
                shutil.move(expected_out_filename + '.log', convert_args.out + '.log')

        # Check if the output is a ped or binary-ped
        elif convert_args.out_format in ['ped', 'ped-12', 'binary-ped']:

            # Call the VCF to PED/BED converter
            convert_vcf_to_plink(convert_args.vcf, convert_args.vcf_fid, convert_args.out_prefix, convert_args.out_format, threads = convert_args.threads, overwrite = convert_args.overwrite)

            # Rename the plink reference
            shutil.move(convert_args.out_prefix + '.log', expected_out_filename + '.log')

            # Add the reference to the log
            log_plink_reference(expected_out_filename, append_mode = True)

    # Check if input is a specified as a ped
    elif convert_args.ped_prefix or (convert_args.ped_filename or convert_args.map_filename):

        # Check that the input and output formats differ
        check_if_conversion('ped', convert_args.out_format)

        # List of formats that are supported from ped
        supported_out_format = ['vcf', 'vcf.gz', 'bcf', 'ped-12', 'binary-ped', 'eigenstrat']

        # Check that the conversion is supported
        check_conversion_support(convert_args.out_format, supported_out_format)

        logging.info('Convertion method assigned')

        # Check if output format is not eigenstrat
        if convert_args.out_format != 'eigenstrat':

            # Convert the ped file using plink
            convert_ped(**vars(convert_args))

            # Rename the plink reference
            shutil.move(convert_args.out_prefix + '.log', expected_out_filename + '.log')

            # Add the reference to the log
            log_plink_reference(expected_out_filename, append_mode = True)

        # Check if output format is eigenstrat
        else:

            # Convert the ped file using admixtools
            ped_to_eigenstrat(**vars(convert_args))

    # Check if input is a specified as a bed
    elif convert_args.bed_prefix or (convert_args.bed_filename or convert_args.fam_filename or convert_args.bim_filename):

        # Check that the input and output formats differ
        check_if_conversion('binary-ped', convert_args.out_format)

        # List of formats that are supported from bed
        supported_out_format = ['vcf', 'vcf.gz', 'bcf', 'ped', 'ped-12', 'eigenstrat']

        # Check that the conversion is supported
        check_conversion_support(convert_args.out_format, supported_out_format)

        logging.info('Convertion method assigned')

        # Check if output format is not eigenstrat
        if convert_args.out_format != 'eigenstrat':

            # Convert the ped file using plink
            convert_bed(**vars(convert_args))

            # Rename the plink reference
            shutil.move(convert_args.out_prefix + '.log', expected_out_filename + '.log')

            # Add the reference to the log
            log_plink_reference(expected_out_filename, append_mode = True)

        # Check if output format is eigenstrat
        else:

            # Convert the ped file using admixtools
            bed_to_eigenstrat(**vars(convert_args))
            

if __name__ == "__main__":
    initLogger()
    run()



#!/usr/bin/env python
'''
    Splits a single VCF file into multiple VCF files using either a statistic or
    bed file. The statistic/bed file must contain loci-based (i.e. window-based)
    data for the function to operate. If the specified statistic file does not
    contain a BIN_END column, the --statistic-window-size argument may be used.

    ############################
    Input Command-line Arguments
    ############################
    **--vcf** *<input_filename>*
        Argument used to define the filename of the VCF file to be split.
    **--split-file** *<split_filename>*
        Argument used to define the file to be split
    **--model-file** *<model_filename>*
        Argument used to define the model file. Please note that this argument cannot be 
        used with the **--pop-file** argument or individual-based filters.
    **--model** *<model_str>*
        Argument used to define the model (i.e. the individual(s) to include and/or the 
        populations for relevant statistics). May be used with any statistic. Please note 
        that this argument cannot be used with **--pop-file** argument or the 
        individual-based filters.

    #############################
    Output Command-line Arguments
    #############################
    **--out-prefix** *<output_prefix>*
        Argument used to define the output prefix (i.e. filename without file extension)
    **--out-format** *<vcf, vcf.gz, bcf>*
        Argument used to define the desired output format. Formats include: uncompressed 
        VCF (vcf); compressed VCF (vcf.gz) [default]; and BCF (bcf).
    **--out-dir** *<output_dir_name>*
        Argument used to define the output directory.
    **--overwrite**
        Argument used to define if previous output should be overwritten.

    ############################
    Split Command-line Arguments
    ############################
    **--split-method** *<statistic-file, bed>*
        Argument used to define the splitting method. Users may spilit using either 
        a statistic-file (statistic-file) from VCF Calc (or other methods) or a BED 
        file (bed).
    **--statistic-window-size** *<statistic_window_int>*
        Argument used to define the size of window calculations. This argument is only 
        required if the BIN_END column is absent within the file.
    **--no-window-correction**
        Argument used to define if a window should not be corrected to avoid 
        an overlap of a single position (i.e. 100-200/200-300 vs. 100-199/200-299).

    #############################
    Filter Command-line Arguments
    #############################
    If using an unfiltered VCF file (e.g. reduce the creation of unnecessary large files)
    the VCF calculator is able to use either a kept or removed sites/BED file and the
    individual-based paramemeters. 
    
    **************************
    Individual-Based Arguments
    **************************
    Please note that all individual-based arguments are not compatible with either the 
    **--model** or **--model-file** command-line arguments.

    **--filter-include-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to include. This argument may be used 
        multiple times if desired.
    **--filter-exclude-indv** *<indv_str>* *<indv1_str, indv2_str, etc.>*
        Argument used to define the individual(s) to exclude. This argument may be used 
        multiple times if desired.
    **--filter-include-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to include.
    **--filter-exclude-indv-file** *<indv_filename>*
        Argument used to define a file of individuals to exclude.
    
    ************************
    Position-Based Arguments
    ************************
    **--filter-include-positions** *<position_filename>*
        Argument used to define a file of positions to include within a tsv file 
        (chromosome and position).
    **--filter-exclude-positions** *<position_filename>*
        Argument used to define a file of positions to exclude within a tsv file 
        (chromosome and position).
    **--filter-include-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to include.
    **--filter-exclude-bed** *<position_bed_filename>*
        Argument used to define a BED file of positions to exclude.

    ##########################
    Example Command-line Usage
    ##########################
    Command-line to split using a statistic file:

    .. code-block:: bash
        
        python vcf_split.py --vcf input.vcf --split-file input.windowed.weir.fst --split-method statistic-file
'''

import os
import sys
import copy
import argparse
import logging
import shutil
import pandas as pd
import numpy as np

from pybedtools import BedTool

# Import basic vcftools functions
from pgpipe.bcftools import *

# Model file related functions
from pgpipe.model import read_model_file

# Import logging module
from pgpipe.logging_module import initLogger, logArgs

# Import vcf format check
from pgpipe.vcf_reader_func import checkFormat

def vcf_split_parser(passed_arguments):
    '''
    VCF Split Argument Parser

    Assign the parameters for VCF Split using argparse.

    Parameters
    ----------
    passed_arguments : list, optional
        Parameters passed by another function. sys.argv is used if
        not given. 

    Raises
    ------
    IOError
        If the input, or other specified files do not exist
    '''

    def parser_confirm_file ():
        '''Custom action to confirm file exists'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError('%s not found' % value)
                setattr(args, self.dest, value)
        return customAction

    def parser_add_to_list ():
        '''Custom action to add items to a list'''
        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):

                # Clean up any commas
                value = [item.strip(',') for item in value]
                
                if not getattr(args, self.dest):
                    setattr(args, self.dest, value)
                else:
                    getattr(args, self.dest).extend(value)
        return customAction

    def metavar_list (var_list):
        '''Create a formmated metavar list for the help output'''
        return '{' + ', '.join(var_list) + '}'

    vcf_parser = argparse.ArgumentParser(formatter_class = argparse.ArgumentDefaultsHelpFormatter)

    # Input arguments.
    vcf_parser.add_argument('--vcf', help = "Defines the filename of the VCF", type = str, required = True, action = parser_confirm_file())

    # Model file arguments.
    vcf_parser.add_argument('--model-file', help = 'Defines the model file', type = str, action = parser_confirm_file())
    vcf_parser.add_argument('--model', help = 'Defines the model and the individual(s) to include', type = str)

    # Split arguments
    vcf_parser.add_argument('--split-file', help='Defines the file to split', required = True, type = str, action = parser_confirm_file())

    split_list = ['statistic-file', 'bed']
    vcf_parser.add_argument('--split-method', metavar = metavar_list(split_list), help = 'Defines the splitting method', type=str, choices = split_list)
    vcf_parser.add_argument('--statistic-window-size', help = 'Defines the size of window calculations (if BIN_END is absent in file)', type = int)
    vcf_parser.add_argument('--no-window-correction', help = 'Defines if the window correction should not be used', action = 'store_true')

    # Output file arguments
    out_format_list = ['vcf', 'vcf.gz', 'bcf']
    out_format_default = 'vcf.gz'
    vcf_parser.add_argument('--out-format', metavar = metavar_list(out_format_list), help = 'Defines the desired output format', type = str, choices = out_format_list, default = out_format_default)
    vcf_parser.add_argument('--out-dir', help = 'Defines the output directory', type = str,  default = 'Split_VCFs')
    vcf_parser.add_argument('--out-prefix', help = 'Defines the output prefix (i.e. filename without file extension)', type = str,  default = 'out')

    # General arguments.
    vcf_parser.add_argument('--overwrite', help = "Overwrite previous output files", action = 'store_true')

    # Galaxy Option to pipe log to stdout
    vcf_parser.add_argument('--log-stdout', help = argparse.SUPPRESS, action = 'store_true')

    ## Filters
    # Position-based position filters
    vcf_parser.add_argument('--filter-include-positions', help = 'Defines a file of positions to include within a tsv file (chromosome and position)', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-positions', help = 'Defines a file of positions to exclude within a tsv file (chromosome and position)', action = parser_confirm_file())

    # BED-based position filters
    vcf_parser.add_argument('--filter-include-bed', help = 'Defines a BED file of positions to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-bed', help = 'Defines a BED file of positions to exclude', action = parser_confirm_file())

    # Non-model based filters
    vcf_parser.add_argument('--filter-include-indv', help = 'Defines the individual(s) to include. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-exclude-indv', help = 'Defines the individual(s) to exclude. This argument may be used multiple times if desired', nargs = '+', type = str, action = parser_add_to_list())
    vcf_parser.add_argument('--filter-include-indv-file', help = 'Defines a file of individuals to include', action = parser_confirm_file())
    vcf_parser.add_argument('--filter-exclude-indv-file', help = 'Defines a file of individuals to exclude', action = parser_confirm_file())

    if passed_arguments:
        return vcf_parser.parse_args(passed_arguments)
    else:
        return vcf_parser.parse_args()

def return_missing_columns (sample_headers):
    '''Reports if a required column (i.e. CHROM, BIN_START, BIN_END) is missing
    from the header of the statistic file.'''

    if not all(pos_header in sample_headers for pos_header in ['CHROM', 'BIN_START', 'BIN_END']):
        # Assign missing headers
        missing_headers = [pos_header for pos_header in ['CHROM', 'BIN_START', 'BIN_END'] if pos_header not in sample_headers]

        return missing_headers

    return None

def split_samples_iter (split_sample_data, split_method):

    # Check if the split method is a statistic file
    if split_method == 'statistic-file':

        for sample_index, split_sample in split_sample_data.iterrows():

            yield sample_index, split_sample

    # Check if the split method is a bed file
    elif split_method == 'bed':

        for sample_index, split_sample in enumerate(split_sample_data):

            yield sample_index, split_sample

def assign_position_args (sample_data, split_method):

    # Assign the statistic file columns
    if split_method == 'statistic-file':

        # Return the position information
        return  str(sample_data['CHROM']), int(sample_data['BIN_START']), int(sample_data['BIN_END'])

    # Assign the BED file columns
    elif split_method == 'bed':

        # Return the position information
        return str(sample_data['chrom']), int(sample_data['start']), int(sample_data['end'])

    else:
        raise Exception('Position assignment error due to split-file format')
 
def run (passed_arguments = []):
    '''
    Split VCF file into multiple VCFs.

    This function uses the argparse-based function :py:func:`vcf_split_parser`
    to parse either sys.argv or passed_arguments to obtain the parameters below. 
    The parameters are then translated to their VCFtools equivalent. Once all 
    the parameters are assigned, bcftools is called.

    Parameters
    ----------
    --vcf : str
        Filename of the VCF
    --out-prefix : str
        Output prefix
    --out-dir : str
        Output directory
    --out-format : str
        Desired output format
    --model-file : str
        Model filename
    --model : str
        Model to use
    --split-file : str
        Filename of file to split
    --split-method : str
        Splitting method
    --statistic-window-size : int
        Size of window calculations (if BIN_END column is absent)
    --no-window-correction : bool
        If the window correction should not be used
    --overwrite : bool
        If previous output should be overwritten    
    --filter-include-indv : list or str
        Individual(s) to include. May be used multiple times. Not usable w/ --model
    --filter-exclude-indv : list or str
        Individual(s) to exclude. May be used multiple times. Not usable w/ --model
    --filter-include-indv-file : str
        File of individuals to include. Not usable w/ --model
    --filter-exclude-indv-file : str
        File of individuals to exclude.  Not usable w/ --model
    --filter-include-positions : str
        File of positions to include (tsv: chromosome and position)
    --filter-exclude-positions : str
        File of positions to exclude (tsv: chromosome and position)
    --filter-include-bed : str
        BED file of positions to include 
    --filter-exclude-bed : str
        BED file of positions to exclude

    Raises
    ------
    IOError
        Output file already exists and --overwrite is not specified
    Exception
        Incompatible arguments
    '''

    # Grab VCF arguments from command line
    vcf_args = vcf_split_parser(passed_arguments)

    # Adds the arguments (i.e. parameters) to the log file
    logArgs(vcf_args, func_name = 'vcf_split')

    # Check if the input file is indexed
    vcf_is_indexed = check_for_index(vcf_args.vcf)

    # Argument container for bcftools
    bcftools_filter_args = ['view']

    # Check if previous output should be overwritten
    if not vcf_args.overwrite:
        # Check if the output directory is present
        if os.path.isdir(vcf_args.out_dir):
            raise IOError('%s already exists' % vcf_args.out_dir)
    else:
        # Check if the output directory is present
        if os.path.isdir(vcf_args.out_dir):
            # Delete the directory
            shutil.rmtree(vcf_args.out_dir)

    # Check if the user has specified a model file
    if vcf_args.model_file and vcf_args.model:
        # Read in the models
        models_in_file = read_model_file(vcf_args.model_file)

        # Check that the selected model was not found in the file
        if vcf_args.model not in models_in_file:
            raise IOError('Selected model "%s" not found in: %s'
                           % (vcf_args.model, vcf_args.model_file))

        # Check that individual-based filters are not also being used
        if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
            if vcf_args.filter_include_indv:
                raise Exception('--model and --filter-include-indv arguments '
                                'are incompatible')
            if vcf_args.filter_exclude_indv:
                raise Exception('--model and --filter-exclude-indv arguments '
                                'are incompatible')

        # Check that individuals-based filters are not also being used
        if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
            if vcf_args.filter_include_indv_file:
                raise Exception('--model and --filter_include_indv-file '
                                'arguments are incompatible')
            if vcf_args.filter_exclude_indv_file:
                raise Exception('--model and --filter-exclude-indv-file '
                                'arguments are incompatible')

        # Select model, might change this in future versions
        selected_model = models_in_file[vcf_args.model]

        # Create individuals file
        selected_model.create_ind_file(overwrite = vcf_args.overwrite)

        # Assign the individuals file to bcftools
        bcftools_filter_args.extend(['--samples-file', selected_model.ind_file])

        logging.info('Model file and parameters assigned')

    # Individuals-based file filters
    if vcf_args.filter_include_indv_file or vcf_args.filter_exclude_indv_file:
        if vcf_args.filter_exclude_indv_file:
            bcftools_filter_args.extend(['--samples-file', vcf_args.filter_include_indv_file])
        if vcf_args.filter_exclude_indv_file:
            bcftools_filter_args.extend(['--samples-file',  '^' + vcf_args.filter_exclude_indv_file])

    # Individual-based filters
    if vcf_args.filter_include_indv or vcf_args.filter_exclude_indv:
        if vcf_args.filter_include_indv:
            bcftools_filter_args.extend(['--samples', ','.join(vcf_args.filter_include_indv)])
        if vcf_args.filter_exclude_indv:
            bcftools_filter_args.extend(['--samples',  '^' + ','.join(vcf_args.filter_exclude_indv)])

    # Position file filters that do not require indexed input
    if (vcf_args.filter_include_positions or vcf_args.filter_exclude_positions) and vcf_is_indexed != True:
        if vcf_args.filter_include_positions:
            bcftools_filter_args.extend(['--targets-file', vcf_args.filter_include_positions])
        if vcf_args.filter_exclude_positions:
            bcftools_filter_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_positions])

    # Position file filters that requires indexed input
    if (vcf_args.filter_include_positions or vcf_args.filter_exclude_positions) and vcf_is_indexed == True:  
        if vcf_args.filter_include_positions:
            bcftools_filter_args.extend(['--regions-file', vcf_args.filter_include_positions])
        if vcf_args.filter_exclude_positions:
            bcftools_filter_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_positions])

    # Position BED filters that do not require indexed input
    if (vcf_args.filter_include_bed or vcf_args.filter_exclude_bed) and vcf_is_indexed != True:
        if vcf_args.filter_include_bed:
            bcftools_filter_args.extend(['--targets-file', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            bcftools_filter_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_bed])

    # Position BED filters that requires indexed input
    if (vcf_args.filter_include_bed or vcf_args.filter_exclude_bed) and vcf_is_indexed == True:  
        if vcf_args.filter_include_bed:
            bcftools_filter_args.extend(['--regions-file', vcf_args.filter_include_bed])
        if vcf_args.filter_exclude_bed:
            bcftools_filter_args.extend(['--targets-file', '^' + vcf_args.filter_exclude_bed])

    logging.info('Filter parameters assigned')

    # Create the vcf/bcf output directory
    if not os.path.exists(vcf_args.out_dir):
        os.makedirs(vcf_args.out_dir)

    # Check if a statistic file is the split method
    if vcf_args.split_method == 'statistic-file':

        # Read in the sample file, make sure the chromosome are read as strings 
        split_samples = pd.read_csv(vcf_args.split_file, sep = '\t', dtype = {'CHROM' : 'str'})

        logging.info('Statistic file for splitting assigned')

        # Assign the columns
        split_columns = list(split_samples)

        # Assign missing columns
        missing_columns = return_missing_columns(split_columns)

        # Check for missing columns
        if missing_columns:

            # Report error if too many or unexpected columns are missing
            if len(missing_columns) != 1 or 'BIN_END' not in missing_columns:
                raise ValueError('Cannot find %s column(s) in file specified by ' \
                                 '--split-file.' % ', '.join(missing_headers))

            # Check if statistic_window_size has been assigned
            if not vcf_args.statistic_window_size:
                raise ValueError("'BIN_END' column absent in %s. Please use " \
                                 '--statistic-window-size' % vcf_args.split_file)

            # Assign window size
            window_size = vcf_args.statistic_window_size

            # Check if the overlap correction has been specified
            if not vcf_args.no_window_correction:

                # Correct the window size
                window_size -= 1

                logging.info('Applied window correction. To disable the correction '
                             "use '--no-window-correction'")

            # Create the 'BIN_END' column
            split_samples['BIN_END'] = split_samples['BIN_START'] + window_size

            logging.info("Calculated 'BIN_END' using --statistic-window-size")

    elif vcf_args.split_method == 'bed':

        # Read in the BED file
        split_samples = BedTool(vcf_args.split_file)

    # Loop the rows of the sample to be split
    for sample_index, split_sample in split_samples_iter(split_samples, vcf_args.split_method):

        # Copy filter arguments for VCF call
        sample_filter_args = copy.deepcopy(bcftools_filter_args)

        # Assign the sample output prefix
        sample_prefix = '%s_%s' % (vcf_args.out_prefix, sample_index)

        # Add the path to the prefix
        sample_path = os.path.join(vcf_args.out_dir, sample_prefix)

        # Assign the position args using the split method
        chromosome, from_bp, to_bp = assign_position_args(split_sample, vcf_args.split_method)

        logging.info('Parameters assigned for locus VCF')

        # Create a single subset VCF for the current loci
        chr_subset_file(vcf_args.vcf, chromosome, sample_path, vcf_args.out_format, from_bp = from_bp, to_bp = to_bp, filter_args = sample_filter_args, overwrite = False)

        logging.info('Locus VCF created')

    # Check if the log should be piped to the stdout
    if vcf_args.log_stdout:

        # Pipe the bcftools reference to stdout
        stdout_bcftools_reference()

    # Check if log should be saved as a file
    else:
    
        # Produce the log file with the bcftools reference
        log_bcftools_reference(vcf_args.out_prefix + '.split')


if __name__ == "__main__":
    initLogger()
    run()



import sys
import argparse
import logging
from pgpipe.genome_region import RegionList, subtractBed

def createParser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--stat-file",dest="stat_file",type=str,help="Name of statistic BED file to be filtered")
    parser.add_argument("--filter-file",dest="filter_file",type=str,help="Name of file with regions that indicate ranges in stat file to be thrown")
    parser.add_argument("--window",type=int,help=("Expand regions in filter file by this number of bases"),default=0)
    parser.add_argument("--zero-ho",action="store_true",dest="zeroho")
    parser.add_argument("--out",type=str)
    return parser

def filter_stat(sysargv):
    parser = createParser()
    args = parser.parse_args(sysargv)
    stat_list = RegionList(filename=args.stat_file,zeroho=args.zeroho,keep_full_line=True)
    filter_list = RegionList(filename=args.filter_file,zeroho=args.zeroho)
    original_len = len(stat_list.regions)
    if args.window != 0:
        filter_list.expandRegions(args.window)
    subtractBed(stat_list,filter_list)
    logging.warning("%d of %d regions selected as non-overlapping"%(len(stat_list.regions),original_len))
    stat_list.printList(file_name=args.out)


if __name__ == "__main__":
    filter_stat(sys.argv[1:])


import sys
import pysam
import argparse
import os.path
import logging
from random import sample

from pgpipe.logging_module import initLogger, logArgs
from pgpipe.genome_region import Region, RegionList
import pgpipe.vcf_reader_func as vf
from pgpipe.parse_functions import defaultsDictForFunction, getConfigFilename, makeRequiredList, getArgsWithConfig
#from tabix_wrapper import prepVcf

#Input: VCF file, reference sequence, region list (possibly .bed file)
#Output: Sequences with reference genome overlayed with VCF SNP calls


def createParser():
    parser = argparse.ArgumentParser(description=("Generates sequences from"
                                     " samples from a VCF file, a reference"
                                     " genome, and a list of gene regions."))
    parser.add_argument("vcfname", help="Input VCF filename")
    parser.add_argument("refname", help="Reference FASTA file")
    parser.add_argument("genename", help="Name of gene region file")
    parser.add_argument("--zero-ho", dest="zeroho", action="store_true")
    parser.add_argument("--zero-closed", dest="zeroclosed", action="store_true")
    parser.add_argument("--indels", dest="indel_flag", action="store_true",
                        help="Include indels when reporting sequences")
    parser.add_argument("--trim-to-ref-length", dest="trim_seq",
                        action="store_true",
                        help=("Trims sequences if indels cause them to be "
                              "longer than reference"))
    parser.add_argument("--output", dest="output_name", help= (
                        "Optional name for output other than default"))
    parser.add_argument("--bed-column-index", dest="gene_col", help= (
                        "Comma-separated list of columns for gene region "
                        " data, format is start/end if no chromosome "
                        " data, start/end/chrom if so"))
    parser.add_argument("--compress-vcf", dest="compress_flag",
                        action="store_true", help=("If input VCF is not "
                        "compressed, will compress and use zip search"))
    parser.add_argument("--conf", dest="config_name", help= ("Name of "
                        "file with configuration options"))
    parser.add_argument("--no-ref-check", dest="ref_check",
                        action="store_false", help=("Prevents exception "
                        "generated by mismatched reference alleles from "
                        "VCF file compared to reference"))
    parser.add_argument("--ima", dest="ima", action="store_true",
                        help=("If set, will output IMa-format instead of "
                              "FASTA format"))
    subsamp_group = parser.add_mutually_exclusive_group()
    subsamp_group.add_argument('--subsamp-list', dest="subsamp_fn",
                               help="List of sample names to be used")
    subsamp_group.add_argument('--subsamp-num', dest="subsamp_num",
                               help=("Number of individuals to be randomly "
                                     "subsampled from VCF file"))
    return parser


def validateFiles(args):
    """Validates that files provided to args all exist on users system"""
    for var in ['vcfname', 'refname', 'genename']:
        f = vars(args)[var]
        if not os.path.exists(f):
            raise ValueError('Filepath for %s not found at %s' %
                            (var, f))


def getMaxAlleleLength(alleles):
    """If an indel, returns length of longest allele (returns 1 for snp)"""
    return max([len(r) for r in alleles])


def getNextIdx(rec, prev_indiv, prev_idx):
    """Using record sample array, find individual and haplotype indices for
    next sample. Will work for any ploidy. Returns -1's when all haplotypes
    have been iterated through"""
    if len(rec.samples[prev_indiv].alleles) > prev_idx + 1:
        return prev_indiv, prev_idx+1
    if len(rec.samples) > prev_indiv+1:
        return prev_indiv+1, 0
    return -1, -1


def checkRefAlign(vcf_recs, fasta_ref, chrom, ref_check):
    """Compares sequence from record to reference FASTA sequence"""
    for vcf_r in vcf_recs:
        vcf_seq = vcf_r.ref
        pos = vcf_r.pos-1
        fasta_seq = fasta_ref.fetch(chrom, pos, pos+len(vcf_seq))
        if vcf_seq != fasta_seq:
            if ref_check:
                raise Exception(("VCF bases and reference bases do not match."
                        "\nVCF reference: %s\nFASTA reference: "
                        "%s\nPosition: %s") % (vcf_seq, fasta_seq, str(pos)))
            else:
                logging.warning("Bases at position %s do not match" %
                            str(pos))


def generateSequence(rec_list, ref_seq, fasta_ref,
                     region, chrom, indiv, idx, args):
    """Fetches variant sites from a given region, then outputs sequences
    from each individual with their correct variants. Will print sequence
    up to the variant site, then print correct variant. After last site,
    will output the rest of the reference sequence."""
    #var_sites = vcf_reader.fetch(chrom,region.start,region.end)
    fl = 0
    seq = ''
    prev_offset = 0

    for vcf_record in rec_list:
        issnp = vf.checkRecordIsSnp(vcf_record)
        if not args.indel_flag and not issnp:
            continue

        pos_offset = vcf_record.pos - 1 - region.start
        for i in range(prev_offset, pos_offset):
            seq += ref_seq[i]
        allele = vcf_record.samples[indiv].alleles[idx]
        if allele is None:
            raise Exception(("Individual %d at position %d is missing "
            "data") % (vcf_record.pos,indiv))
        if issnp:
            seq += vcf_record.samples[indiv].alleles[idx]
            prev_offset = pos_offset+1
        else:
            max_indel = getMaxAlleleLength(vcf_record.alleles)
            allele = vcf_record.samples[indiv].alleles[idx]
            for i in range(len(allele), max_indel):
                allele += '_'
            seq += allele
            indel_offset = len(vcf_record.ref)
            prev_offset = pos_offset+indel_offset

    for i in range(prev_offset, len(ref_seq)):
        seq += ref_seq[i]

    if args.trim_seq:
        return seq[:len(ref_seq)]
    return seq

def getHeader(record_count, chrom, region, zeroho=False, zeroclosed=False):
    start = region.start
    end = region.end
    if not zeroho:
        start += 1
    elif zeroclosed:
        end -= 1
    return '>'+str(record_count)+' '+chrom+' '+str(start)+':'+str(end)


def getFastaFilename(args):
    vcfname = args.vcfname
    for ext in ['vcf.gz', 'vcf', 'bcf', 'vcf.bgz']:
        if ext == vcfname[-1*len(ext):]:
            if args.output_name is None:
                return vcfname[:-1*len(ext)]+'fasta'
            else:
                return args.output_name
    raise Exception('VCF filename %s has no valid extension' %
                    vcfname)


def vcf_to_seq(sys_args):
    """Returns a FASTA file with seqs from individuals in given gene regions

    Given an input VCF file, a reference FASTA file, and a list of gene
    regions, will output a FASTA file with sequence data for all individuals
    in the regions given. The reference FASTA file must be a full file from
    one or multiple chromosomes, starting at the first base. The gene
    region file must have start and end coordinates (half-open), with an
    optional column for chromosome data if the VCF input has multiple
    chromosomes.

    Parameters
    ----------
    --vcf : str
        Filename for VCF input file. If it does not end with extension
        'vcf(.gz)', a value for --ext must be provided.
    --ref : str
        Filename for FASTA reference file. This file can contain multiple
        chromosomes but must start from the first base, as there is currently
        no way to offset the sequences when pulling from a Region
    --rl : str
        Filename for gene region file. Requires columns for start and end
        coordinates, with option for chromosome. Additional data may be
        included, the columns with relevant data can be specified with the
        --gene-col option
    --indels : bool, optional
        If set, indels will be included in the output sequences
    --output : str, optional
        If set, the default output name of (inputprefix).fasta will be
        replaced with the given string
    --gene-col : str, optional
        Comma-separated string with two or three elements (chromosome is
        optional). If length is 2, elements are the indices for columns
        in the input gene region file corresponding to the start/end
        coordinates of a region. If length 3, the third element
        specifies the index of the chromosome column. Default is "1,2,0",
        to match column order in a BED file.



    Other Parameters
    ----------------
    --gr1 : bool, optional (False)
        If set, indicates that the genome coordinate data is in base 1
    --trim-to-ref-length: bool, optional (False)
        If set, the sequences output will always match the length of the
        region they are found in. For example, a sequence with an insertion
        will cause the sequence to be an additional length of n-1, with n
        being the length of the insertion.
    --compress-vcf : bool, optional
        If set, will use bgzip and tabix to compress and index given VCF
        file
    --subsamp-list : str, optional
        Name of single-column file with names of individuals to subsample
        from input VCF file.
    --subsamp-num : int, optional
        Number of individuals to be randomly subsampled from VCF file

    Output
    ------
    FASTA file
        Will be named either '--output' value or (vcfinput).fasta.
        Contains full sequence for given regions and individuals with
        appropriate SNP/indel data included.
    """
    parser = createParser()
    if len(sys_args) == 0:
        parser.print_help()
        sys.exit(1)

    required_args = ['vcfname','refname','genename']
    args = getArgsWithConfig(parser,sys_args,required_args,'vcf_ref_to_seq')

    logArgs(args)
    validateFiles(args)
    if args.output_name is not None:
        fasta_filename = args.output_name
    else:
        fasta_filename = getFastaFilename(args)
    fasta_file = open(fasta_filename, 'w')

    vcf_reader, uncompressed = vf.getVcfReader(args.vcfname,
                               compress_flag=args.compress_flag,
                               subsamp_num=args.subsamp_num,
                               subsamp_fn=args.subsamp_fn)
    logging.info('VCF file read')
    first_el = next(vcf_reader)
    chrom = first_el.chrom
    #compressed = (input_ext != 'vcf')

    region_list = RegionList(filename=args.genename, zeroho=args.zeroho,
                            zeroclosed=args.zeroclosed,
                            colstr=args.gene_col)
    logging.info('Region list read')
    fasta_ref = pysam.FastaFile(args.refname)
    record_count = 1
    prev_last_rec = first_el
    logging.info('Total individuals: %d' % (len(first_el.samples)))
    logging.info('Total regions: %d' % (len(region_list.regions)))
    for region in region_list.regions:
        if not uncompressed:
            rec_list = vf.getRecordList(vcf_reader, region)
        else:
            rec_list, prev_last_rec = vf.getRecordListUnzipped(vcf_reader,
                                      prev_last_rec, region)
        if len(rec_list) == 0:
            logging.warning(("Region from %d to %d has no variants "
                            "in VCF file") % (region.start,region.end))
        logging.debug('Region %d to %d: %d variants' %
                      (region.start,region.end,len(rec_list)))
        ref_seq = fasta_ref.fetch(chrom, region.start, region.end)
        checkRefAlign(rec_list, fasta_ref, chrom, args.ref_check)
        fasta_header = getHeader(record_count, chrom, region,
                                 zeroho=args.zeroho, zeroclosed=args.zeroclosed)
        fasta_file.write(fasta_header+'\n')
        indiv, idx = 0, 0
        while indiv != -1:
            seq = generateSequence(rec_list, ref_seq, fasta_ref,
                                   region, chrom, indiv, idx, args)
            fasta_file.write(seq+'\n')
            indiv, idx = getNextIdx(first_el, indiv, idx)
        record_count += 1
    fasta_file.close()

if __name__ == "__main__":
    vcf_to_seq(sys.argv[1:])
